"""Module to handle image quality calculations."""
#
# iqcalc.py -- image quality calculations on FITS data
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

import math
import logging
import threading

import numpy as np

try:
    import scipy.optimize as optimize
    import scipy.ndimage as ndimage
    from scipy.ndimage import maximum_filter
    from scipy.interpolate import interp1d
    have_scipy = True
except ImportError:
    have_scipy = False

from ginga.misc import Bunch

__all__ = ['get_mean', 'get_median', 'IQCalcError', 'IQCalc']


def get_mean(data_np):
    """Calculate mean for valid values.

    Parameters
    ----------
    data_np : ndarray
        Input array. Can contain masked values.

    Returns
    -------
    result : float
        Mean of array values that are finite.
        If array contains no finite values, returns NaN.

    """
    i = np.isfinite(data_np)
    if not np.any(i):
        return np.nan
    # NOTE: we use "ma" version of mean because this can be used with
    # masked arrays created by cutting out non-rectangular shapes
    return np.ma.mean(data_np[i])


def get_median(data_np):
    """Like :func:`get_mean` but for median."""
    i = np.isfinite(data_np)
    if not np.any(i):
        return np.nan
    # NOTE: we use "ma" version of median because this can be used with
    # masked arrays created by cutting out non-rectangular shapes
    return np.ma.median(data_np[i])


class IQCalcError(Exception):
    """Base exception for raising errors in this module."""
    pass


class IQCalc(object):
    """Class to handle model fitting and FWHM calculations.

    Parameters
    ----------
    logger : obj or `None`
        Python logger. If not given, one will be created.

    Attributes
    ----------
    lock : :py:class:`threading.RLock`
        For mutex around `scipy.optimize`, which seems to be non-threadsafe.

    skylevel_magnification, skylevel_offset : float
        For adjustments to sky background level.

    """
    def __init__(self, logger=None):
        if not logger:
            logger = logging.getLogger('IQCalc')
        self.logger = logger

        # for mutex around scipy.optimize, which seems to be non-threadsafe
        self.lock = threading.RLock()

        # for adjustments to background level
        self.skylevel_magnification = 1.05
        self.skylevel_offset = 40.0

    # FWHM CALCULATION

    def gaussian(self, x, p):
        """Evaluate Gaussian function in 1D. See :meth:`calc_fwhm`.

        Parameters
        ----------
        x : array-like
            X values.

        p : tuple of float
            Parameters for Gaussian, i.e., ``(mean, stddev, amplitude)``.

        Returns
        -------
        y : array-like
            Y values.

        """
        y = (1.0 / (p[1] * np.sqrt(2 * np.pi)) *
             np.exp(-(x - p[0]) ** 2 / (2 * p[1] ** 2))) * p[2]
        return y

    def calc_fwhm_gaussian(self, arr1d, medv=None, gauss_fn=None):
        """FWHM calculation on a 1D array by using least square fitting of
        a Gaussian function on the data.

        Parameters
        ----------
        arr1d : array-like
            1D array cut in either X or Y direction on the object.

        medv : float or `None`
            Median of the data. If not given, it is calculated from ``arr1d``.

        gauss_fn : func or `None`
            Gaussian function for fitting. If not given, :meth:`gaussian`
            is used.

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            Fitting results.

        Raises
        ------
        IQCalcError
            Fitting failed.

        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        if gauss_fn is None:
            gauss_fn = self.gaussian

        N = len(arr1d)
        X = np.array(list(range(N)))
        Y = arr1d
        # Fitting works more reliably if we do the following
        # a. subtract sky background
        if medv is None:
            medv = get_median(Y)
        Y = Y - medv
        maxv = Y.max()
        # b. clamp to 0..max (of the sky subtracted field)
        Y = Y.clip(0, maxv)

        # Fit a gaussian
        p0 = [0, N - 1, maxv]              # Inital guess
        # Distance to the target function
        errfunc = lambda p, x, y: gauss_fn(x, p) - y  # noqa
        # Least square fit to the gaussian
        with self.lock:
            # NOTE: without this mutex, optimize.leastsq causes a fatal error
            # sometimes--it appears not to be thread safe.
            # The error is:
            # "SystemError: null argument to internal routine"
            # "Fatal Python error: GC object already tracked"
            p1, success = optimize.leastsq(errfunc, p0[:], args=(X, Y))

        if not success:
            raise IQCalcError("FWHM Gaussian fitting failed")

        mu, sdev, maxv = p1
        self.logger.debug("mu=%f sdev=%f maxv=%f" % (mu, sdev, maxv))

        # Now that we have the sdev from fitting, we can calculate FWHM
        fwhm = 2.0 * np.sqrt(2.0 * np.log(2.0)) * sdev
        # some routines choke on numpy values and need "pure" Python floats
        # e.g. when marshalling through a remote procedure interface
        fwhm = float(fwhm)
        mu = float(mu)
        sdev = float(sdev)
        maxv = float(maxv)

        res = Bunch.Bunch(fwhm=fwhm, mu=mu, sdev=sdev, maxv=maxv,
                          fit_fn=gauss_fn, fit_args=[mu, sdev, maxv])
        return res

    def moffat(self, x, p):
        """Evaluate Moffat function in 1D. See :meth:`calc_fwhm`.

        Parameters
        ----------
        x : array-like
            X values.

        p : tuple of float
            Parameters for Moffat, i.e., ``(x_0, gamma, alpha, amplitude)``,
            where ``x_0`` a.k.a. mean and ``gamma`` core width.

        Returns
        -------
        y : array-like
            Y values.

        """
        y = (1.0 + (x - p[0]) ** 2 / p[1] ** 2) ** (-1.0 * p[2]) * p[3]
        return y

    def calc_fwhm_moffat(self, arr1d, medv=None, moffat_fn=None):
        """FWHM calculation on a 1D array by using least square fitting of
        a Moffat function on the data.

        Parameters
        ----------
        arr1d : array-like
            1D array cut in either X or Y direction on the object.

        medv : float or `None`
            Median of the data. If not given, it is calculated from ``arr1d``.

        moffat_fn : func or `None`
            Moffat function for fitting. If not given, :meth:`moffat` is used.

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            Fitting results.

        Raises
        ------
        IQCalcError
            Fitting failed.

        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        if moffat_fn is None:
            moffat_fn = self.moffat

        N = len(arr1d)
        X = np.array(list(range(N)))
        Y = arr1d
        # Fitting works more reliably if we do the following
        # a. subtract sky background
        if medv is None:
            medv = get_median(Y)
        Y = Y - medv
        maxv = Y.max()
        # b. clamp to 0..max (of the sky subtracted field)
        Y = Y.clip(0, maxv)

        # Fit a moffat
        p0 = [0, N - 1, 2, maxv]              # Inital guess
        # Distance to the target function
        errfunc = lambda p, x, y: moffat_fn(x, p) - y  # noqa
        # Least square fit to the gaussian
        with self.lock:
            # NOTE: without this mutex, optimize.leastsq causes a fatal error
            # sometimes--it appears not to be thread safe.
            # The error is:
            # "SystemError: null argument to internal routine"
            # "Fatal Python error: GC object already tracked"
            p1, success = optimize.leastsq(errfunc, p0[:], args=(X, Y))

        if not success:
            raise IQCalcError("FWHM Moffat fitting failed")

        mu, width, power, maxv = p1
        width = np.abs(width)
        self.logger.debug("mu=%f width=%f power=%f maxv=%f" % (
            mu, width, power, maxv))

        fwhm = 2.0 * width * np.sqrt(2.0 ** (1.0 / power) - 1.0)

        # some routines choke on numpy values and need "pure" Python floats
        # e.g. when marshalling through a remote procedure interface
        fwhm = float(fwhm)
        mu = float(mu)
        width = float(width)
        power = float(power)
        maxv = float(maxv)

        res = Bunch.Bunch(fwhm=fwhm, mu=mu, width=width, power=power,
                          maxv=maxv, fit_fn=moffat_fn,
                          fit_args=[mu, width, power, maxv])
        return res

    def calc_fwhm(self, arr1d, medv=None, method_name='gaussian'):
        """Calculate FWHM for the given input array.

        Parameters
        ----------
        arr1d : array-like
            1D array cut in either X or Y direction on the object.

        medv : float or `None`
            Median of the data. If not given, it is calculated from ``arr1d``.

        method_name : {'gaussian', 'moffat'}
            Function to use for fitting.

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            Fitting results.

        """
        # Calculate FWHM in each direction
        fwhm_fn = self.calc_fwhm_gaussian
        if method_name == 'moffat':
            fwhm_fn = self.calc_fwhm_moffat

        return fwhm_fn(arr1d, medv=medv)

    def get_fwhm(self, x, y, radius, data, medv=None, method_name='gaussian'):
        """Get the FWHM values of the object at the given coordinates and
        radius.

        Parameters
        ----------
        x, y : int
            Indices of the object location in data array.

        radius : float
            Radius of the region encompassing the object.

        data : array-like
            Data array.

        medv, method_name
            See :meth:`calc_fwhm`.

        Returns
        -------
        fwhm_x, fwhm_y : float
            FWHM in X and Y, respectively.

        ctr_x, ctr_y : float
            Center in X and Y, respectively.

        x_res, y_res : dict
            Fit results from :meth:`calc_fwhm` in X and Y, respectively.

        """
        if medv is None:
            medv = get_median(data)

        # Get two cuts of the data, one in X and one in Y
        x0, y0, xarr, yarr = self.cut_cross(x, y, radius, data)

        # Calculate FWHM in each direction
        x_res = self.calc_fwhm(xarr, medv=medv, method_name=method_name)
        fwhm_x, cx = x_res.fwhm, x_res.mu

        y_res = self.calc_fwhm(yarr, medv=medv, method_name=method_name)
        fwhm_y, cy = y_res.fwhm, y_res.mu

        ctr_x = x0 + cx
        ctr_y = y0 + cy
        self.logger.debug("fwhm_x,fwhm_y=%f,%f center=%f,%f" % (
            fwhm_x, fwhm_y, ctr_x, ctr_y))
        return (fwhm_x, fwhm_y, ctr_x, ctr_y, x_res, y_res)

    def starsize(self, fwhm_x, deg_pix_x, fwhm_y, deg_pix_y):
        """Calculate average FWHM in arcseconds.

        Parameters
        ----------
        fwhm_x : float
            FWHM in X (pixels).

        deg_pix_x : float
            Plate scale from CDELT1 in degrees per pixel.

        fwhm_y : float
            FWHM in Y (pixels).

        deg_pix_y : float
            Plate scale from CDELT2 in degrees per pixel.

        Returns
        -------
        fwhm : float
            Average FWHM in arcseconds.

        """
        cdelta1 = math.fabs(deg_pix_x)
        cdelta2 = math.fabs(deg_pix_y)
        fwhm = (fwhm_x * cdelta1 + fwhm_y * cdelta2) / 2.0
        fwhm = fwhm * 3600.0
        return fwhm

    def centroid(self, data, xc, yc, radius):
        """Calculate centroid from center of mass.

        Parameters
        ----------
        data : array-like
            Data array.

        xc, yc : int
            X and Y indices of the approximate center.

        radius : float
            Half-width of the region to consider around the given center.

        Returns
        -------
        x, y : float
            Centroid indices.

        Raises
        ------
        IQCalcError
            Missing dependency.

        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        xc, yc = int(xc), int(yc)
        x0, y0, arr = self.cut_region(xc, yc, int(radius), data)
        # See https://stackoverflow.com/questions/25369982/center-of-mass-for-roi-in-python
        cp_arr = np.asarray(arr)
        cy, cx = ndimage.center_of_mass(cp_arr)
        return (x0 + cx, y0 + cy)

    # FINDING BRIGHT PEAKS

    def get_threshold(self, data, sigma=5.0):
        """Calculate threshold for :meth:`find_bright_peaks`.

        Parameters
        ----------
        data : array-like
            Data array.

        sigma : float
            Sigma for the threshold.

        Returns
        -------
        threshold : float
            Threshold based on good data, its median, and the given sigma.

        """
        # remove masked elements
        fdata = data[np.logical_not(np.ma.getmaskarray(data))]
        # remove Inf or NaN
        fdata = fdata[np.isfinite(fdata)]

        # find the median
        median = get_median(fdata)

        # NOTE: for this method a good default sigma is 5.0
        dist = np.fabs(fdata - median).mean()
        threshold = median + sigma * dist

        # NOTE: for this method a good default sigma is 2.0
        ## std = np.std(fdata - median)
        ## threshold = median + sigma * std

        self.logger.debug("calc threshold=%f" % (threshold))
        return threshold

    def find_bright_peaks(self, data, threshold=None, sigma=5, radius=5):
        """Find bright peak candidates in in the given data.

        Parameters
        ----------
        data : array-like
            Input data to find peaks from.

        threshold : float or `None`
            Detection threshold. Below this value, an object is not
            considered a candidate. If not given, a default is calculated
            using :meth:`get_threshold` with the given ``sigma``.

        sigma : float
            Sigma for the threshold.

        radius : float
            Pixel radius for determining local maxima. If the
            desired objects are larger in size, specify a larger radius.

        Returns
        -------
        peaks : list of tuple
            A list of candidate object coordinate tuples ``(x, y)`` in data.

        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        if threshold is None:
            # set threshold to default if none provided
            threshold = self.get_threshold(data, sigma=sigma)
            self.logger.debug("threshold defaults to %f (sigma=%f)" % (
                threshold, sigma))

        #self.logger.debug("filtering")
        data_max = maximum_filter(data, radius)
        maxima = (data == data_max)
        diff = data_max > threshold
        maxima[diff == 0] = 0

        #self.logger.debug("finding")
        labeled, num_objects = ndimage.label(maxima)
        slices = ndimage.find_objects(labeled)
        peaks = []
        for dy, dx in slices:
            xc = (dx.start + dx.stop - 1) / 2.0
            yc = (dy.start + dy.stop - 1) / 2.0

            # This is only an approximate center; use FWHM or centroid
            # calculation to refine further
            peaks.append((xc, yc))

        self.logger.debug("peaks=%s" % (str(peaks)))
        return peaks

    def cut_region(self, x, y, radius, data):
        """Return a cut region.

        Parameters
        ----------
        x, y : int
            Indices of central pixel.

        radius : int
            Half-width in both X and Y directions.

        data : array-like
            Data array to cut from.

        Returns
        -------
        x0, y0 : int
            Origin of the region.

        arr : array-like
            Cut region (a view, not copy).

        """
        n = radius
        ht, wd = data.shape
        x0, x1 = max(0, x - n), min(wd - 1, x + n)
        y0, y1 = max(0, y - n), min(ht - 1, y + n)
        arr = data[y0:y1 + 1, x0:x1 + 1]
        return (x0, y0, arr)

    def cut_cross(self, x, y, radius, data):
        """Cut data vertically and horizontally at the given position
        with the given radius.

        Parameters
        ----------
        x, y : int
            Indices where vertical and horizontal cuts meet.

        radius : float
            Radius of both cuts.

        data : array-like
            Data array to cut from.

        Returns
        -------
        x0 : array-like
            Starting pixel of horizontal cut (in X).

        y0 : array-like
            Starting pixel of vertical cut (in Y).

        xarr : array-like
            Horizontal cut (in X).

        yarr : array-like
            Vertical cut (in Y).

        """
        n = int(round(radius))
        ht, wd = data.shape
        x, y = int(round(x)), int(round(y))
        x0, x1 = int(max(0, x - n)), int(min(wd - 1, x + n))
        y0, y1 = int(max(0, y - n)), int(min(ht - 1, y + n))
        xarr = data[y, x0:x1 + 1]
        yarr = data[y0:y1 + 1, x]
        return (x0, y0, xarr, yarr)

    def brightness(self, x, y, radius, medv, data):
        """Return the brightness value found in a region defined by input
        location and radius. Region is cut using :meth:`cut_region`.

        Parameters
        ----------
        x, y : int
            Indices of central pixel.

        radius : int
            Half-width in both X and Y directions.

        medv : float
            Background to subtract off.

        data : array-like
            Data array.

        Returns
        -------
        res : float
            Brightness.

        """
        x0, y0, arr = self.cut_region(x, y, radius, data)
        arr2 = np.sort(arr.flat)
        idx = int(len(arr2) * 0.8)
        res = arr2[idx] - medv
        return float(res)

    def fwhm_data(self, x, y, data, radius=15, method_name='gaussian'):
        """Equivalent to :meth:`get_fwhm`."""
        return self.get_fwhm(x, y, radius, data, method_name=method_name)

    # Encircled and ensquared energies (EE)

    def ensquared_energy(self, data):
        """Return a function of ensquared energy across pixel indices.

        Ideally, data is already a masked array and is assumed to be centered.

        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")

        tot = data.sum()
        ny, nx = data.shape
        cen_x = int(nx // 2)
        cen_y = int(ny // 2)
        ee = []

        if ny > nx:
            n_max = ny
            cen = cen_y
        else:
            n_max = nx
            cen = cen_x

        if n_max % 2 == 0:  # Even
            delta_i1 = -1
        else:  # Odd
            delta_i1 = 0

        xr = range(n_max - cen)

        for i in xr:
            ix1 = cen_x - i + delta_i1
            if ix1 < 0:
                ix1 = 0
            ix2 = cen_x + i + 1
            if ix2 > nx:
                ix2 = nx
            iy1 = cen_y - i + delta_i1
            if iy1 < 0:
                iy1 = 0
            iy2 = cen_y + i + 1
            if iy2 > ny:
                iy2 = ny
            ee.append(data[iy1:iy2, ix1:ix2].sum() / tot)

        return interp1d(xr, ee, kind='cubic', bounds_error=False,
                        assume_sorted=True)

    # This is adapted from poppy package. See licenses/POPPY_LICENSE.md .
    def encircled_energy(self, data):
        """Return a function of encircled energy across pixel indices.

        Ideally, data is already a masked array and is assumed to be centered.

        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")

        y, x = np.indices(data.shape, dtype=float)
        cen = tuple((i - 1) * 0.5 for i in data.shape[::-1])
        x -= cen[0]
        y -= cen[1]
        r = np.sqrt(x * x + y * y)

        ind = np.argsort(r.flat)
        sorted_r = r.flat[ind]
        sorted_data = data.flat[ind]

        # data is already masked
        csim = sorted_data.cumsum(dtype=float)

        sorted_r_int = sorted_r.astype(int)
        deltar = sorted_r_int[1:] - sorted_r_int[:-1]  # assume all radii represented
        rind = np.where(deltar)[0]

        ee = csim[rind] / sorted_data.sum()  # Normalize
        if isinstance(ee, np.ma.MaskedArray):
            ee.set_fill_value(0)
            ee = ee.filled()

        return interp1d(range(ee.size), ee, kind='cubic', bounds_error=False,
                        assume_sorted=True)

    # EVALUATION ON A FIELD

    def evaluate_peaks(self, peaks, data, fwhm_radius=15,
                       fwhm_method='gaussian', ee_total_radius=10,
                       cb_fn=None, ev_intr=None):
        """Evaluate photometry for given peaks in data array.

        Parameters
        ----------
        peaks : list of tuple
            List of ``(x, y)`` tuples containing indices of peaks.

        data : array-like
            Data array that goes with the given peaks.

        fwhm_radius, fwhm_method
            See :meth:`get_fwhm`.

        ee_total_radius : float
            Radius, in pixels, where encircled and ensquared energy fractions
            are defined as 1.

        cb_fn : func or `None`
            If applicable, provide a callback function that takes a
            `ginga.misc.Bunch.Bunch` containing the result for each peak.
            It should not return anything.

        ev_intr : :py:class:`threading.Event` or `None`
            For threading, if applicable.

        .. note:: unused parameter `bright_radius` was removed in
                  release 4.0

        Returns
        -------
        objlist : list of `ginga.misc.Bunch.Bunch`
            A list of successful results for the given peaks.
            Each result contains the following keys:

            * ``objx``, ``objy``: Fitted centroid from :meth:`get_fwhm`.
            * ``pos``: A measure of distance from the center of the image.
            * ``oid_x``, ``oid_y``: Center-of-mass centroid from :meth:`centroid`.
            * ``fwhm_x``, ``fwhm_y``: Fitted FWHM from :meth:`get_fwhm`.
            * ``fwhm``: Overall measure of fwhm as a single value.
            * ``fwhm_radius``: Input FWHM radius.
            * ``brightness``: Average peak value based on :meth:`get_fwhm` fits.
            * ``elipse``: A measure of ellipticity.
            * ``x``, ``y``: Input indices of the peak.
            * ``skylevel``: Sky level estimated from median of data array and
              ``skylevel_magnification`` and ``skylevel_offset`` attributes.
            * ``background``: Median of the input array.
            * ``ensquared_energy_fn``: Function of ensquared energy for different pixel radii.
            * ``encircled_energy_fn``: Function of encircled energy for different pixel radii.

        """
        height, width = data.shape
        hh = float(height) / 2.0
        ht = float(height)
        h4 = float(height) * 4.0
        wh = float(width) / 2.0
        wd = float(width)
        w4 = float(width) * 4.0

        # Find the median (sky/background) level
        median = float(get_median(data))
        #skylevel = median
        # Old SOSS qualsize() applied this calculation to skylevel
        skylevel = median * self.skylevel_magnification + self.skylevel_offset

        # Form a list of objects and their characteristics
        objlist = []
        for x, y in peaks:
            if ev_intr and ev_intr.is_set():
                raise IQCalcError("Evaluation interrupted!")

            # centroid calculation on local peak
            oid_x, oid_y = None, None
            try:
                oid_x, oid_y = self.centroid(data, x, y, fwhm_radius)

            except Exception as e:
                # Error doing centroid
                self.logger.debug("Error doing centroid on object at %.2f,%.2f: %s" % (
                    x, y, str(e)))

            # Find the fwhm in x and y, using local peak
            try:
                res = self.fwhm_data(x, y, data, radius=fwhm_radius,
                                     method_name=fwhm_method)
                fwhm_x, fwhm_y, ctr_x, ctr_y, x_res, y_res = res

                bx = x_res.fit_fn(round(ctr_x),
                                  (ctr_x,) + tuple(x_res.fit_args[1:]))
                by = y_res.fit_fn(round(ctr_y),
                                  (ctr_y,) + tuple(y_res.fit_args[1:]))
                bright = float((bx + by) / 2.0)

            except Exception as e:
                # Error doing FWHM, skip this object
                self.logger.debug("Error doing FWHM on object at %.2f,%.2f: %s" % (
                    x, y, str(e)))
                continue

            self.logger.debug("orig=%f,%f  ctr=%f,%f  fwhm=%f,%f bright=%f" % (
                x, y, ctr_x, ctr_y, fwhm_x, fwhm_y, bright))
            # overall measure of fwhm as a single value
            fwhm = (math.sqrt(fwhm_x * fwhm_x + fwhm_y * fwhm_y) *
                    (1.0 / math.sqrt(2.0)))

            # calculate a measure of ellipticity
            elipse = math.fabs(min(fwhm_x, fwhm_y) / max(fwhm_x, fwhm_y))

            # calculate a measure of distance from center of image
            dx = wh - ctr_x
            dy = hh - ctr_y
            dx2 = dx * dx / wd / w4
            dy2 = dy * dy / ht / h4
            if dx2 > dy2:
                pos = 1.0 - dx2
            else:
                pos = 1.0 - dy2

            # EE on background subtracted image
            ee_sq_fn = None
            ee_circ_fn = None
            iy1 = int(ctr_y - ee_total_radius)
            iy2 = int(ctr_y + ee_total_radius) + 1
            ix1 = int(ctr_x - ee_total_radius)
            ix2 = int(ctr_x + ee_total_radius) + 1

            if iy1 < 0 or iy2 > height or ix1 < 0 or ix2 > width:
                self.logger.debug("Error calculating EE on object at %.2f,%.2f: Box out of range with radius=%.2f" % (x, y, ee_total_radius))
            else:
                ee_data = data[iy1:iy2, ix1:ix2] - median
                try:
                    ee_sq_fn = self.ensquared_energy(ee_data)
                except Exception as e:
                    self.logger.debug("Error calculating ensquared energy on object at %.2f,%.2f: %s" % (x, y, str(e)))
                try:
                    ee_circ_fn = self.encircled_energy(ee_data)
                except Exception as e:
                    self.logger.debug("Error calculating encircled energy on object at %.2f,%.2f: %s" % (x, y, str(e)))

            obj = Bunch.Bunch(objx=ctr_x, objy=ctr_y, pos=pos,
                              oid_x=oid_x, oid_y=oid_y,
                              fwhm_x=fwhm_x, fwhm_y=fwhm_y,
                              fwhm=fwhm, fwhm_radius=fwhm_radius,
                              brightness=bright, elipse=elipse,
                              x=int(x), y=int(y),
                              skylevel=skylevel, background=median,
                              ensquared_energy_fn=ee_sq_fn,
                              encircled_energy_fn=ee_circ_fn)
            objlist.append(obj)

            if cb_fn is not None:
                cb_fn(obj)

        return objlist

    def _sortkey(self, obj):
        """For sorting of result in :meth:`objlist_select`."""
        val = obj.brightness * obj.pos / math.sqrt(obj.fwhm)
        return val

    def objlist_select(self, objlist, width, height,
                       minfwhm=2.0, maxfwhm=150.0, minelipse=0.5,
                       edgew=0.01):
        """Filter output from :meth:`evaluate_peaks`.

        Parameters
        ----------
        objlist : list of `ginga.misc.Bunch.Bunch`
            Output from :meth:`evaluate_peaks`.

        width, height : int
            Dimension of data array from which ``objlist`` was derived.

        minfwhm, maxfwhm : float
            Limits for desired FWHM, where ``(minfwhm, maxfwhm)``.

        minelipse : float
            Minimum value of desired ellipticity (not inclusive).

        edgew : float
            Factor between 0 and 1 that determines if a location is too close to the edge or not.

        Returns
        -------
        results : list of `ginga.misc.Bunch.Bunch`
            Elements of ``objlist`` that contain desired FWHM, ellipticity,
            and not too close to the edge.

        """
        results = []
        count = 0
        for obj in objlist:
            count += 1
            self.logger.debug("%d obj x,y=%.2f,%.2f fwhm=%.2f bright=%.2f" % (
                count, obj.objx, obj.objy, obj.fwhm, obj.brightness))
            # If peak has a minfwhm < fwhm < maxfwhm and the object
            # is inside the frame by edgew pct
            if ((minfwhm < obj.fwhm) and (obj.fwhm < maxfwhm) and
                    (minelipse < obj.elipse) and (width * edgew < obj.x) and
                    (height * edgew < obj.y) and
                    (width * (1.0 - edgew) > obj.x) and
                    (height * (1.0 - edgew) > obj.y)):
                results.append(obj)

        #results.sort(cmp=self._compare)
        results.sort(key=self._sortkey, reverse=True)
        return results

    def pick_field(self, data, peak_radius=5, fwhm_radius=15,
                   threshold=None,
                   minfwhm=2.0, maxfwhm=50.0, minelipse=0.5,
                   edgew=0.01, ee_total_radius=10):
        """Pick the first good object within the given field.

        Parameters
        ----------
        data : array-like
            Data array of the field.

        peak_radius, threshold
            See :meth:`find_bright_peaks`.

        fwhm_radius, ee_total_radius
            See :meth:`evaluate_peaks`.

        minfwhm, maxfwhm, minelipse, edgew
            See :meth:`objlist_select`.

        .. note:: unused parameter `bright_radius` was removed in
                  release 4.0

        Returns
        -------
        result : `ginga.misc.Bunch.Bunch`
            This is a single element of ``objlist`` as described in
            :meth:`evaluate_peaks`.

        Raises
        ------
        IQCalcError
            No object matches selection criteria.

        """
        height, width = data.shape

        # Find the bright peaks in the image
        peaks = self.find_bright_peaks(data, radius=peak_radius,
                                       threshold=threshold)
        self.logger.debug("peaks=%s" % str(peaks))
        if len(peaks) == 0:
            raise IQCalcError("Cannot find bright peaks")

        # Evaluate those peaks
        objlist = self.evaluate_peaks(peaks, data,
                                      fwhm_radius=fwhm_radius,
                                      ee_total_radius=ee_total_radius)
        if len(objlist) == 0:
            raise IQCalcError("Error evaluating bright peaks")

        results = self.objlist_select(objlist, width, height,
                                      minfwhm=minfwhm, maxfwhm=maxfwhm,
                                      minelipse=minelipse, edgew=edgew)
        if len(results) == 0:
            raise IQCalcError("No object matches selection criteria")

        return results[0]

    def qualsize(self, image, x1=None, y1=None, x2=None, y2=None,
                 radius=5, fwhm_radius=15, threshold=None,
                 minfwhm=2.0, maxfwhm=50.0, minelipse=0.5,
                 edgew=0.01, ee_total_radius=10):
        """Run :meth:`pick_field` on the given image.

        Parameters
        ----------
        image : `ginga.AstroImage.AstroImage`
            Image to process.

        x1, y1, x2, y2 : int
            See :meth:`ginga.BaseImage.BaseImage.cutout_data`.

        radius, threshold
            See :meth:`find_bright_peaks`.

        fwhm_radius, ee_total_radius
            See :meth:`evaluate_peaks`.

        minfwhm, maxfwhm, minelipse, edgew
            See :meth:`objlist_select`.

        .. note:: unused parameter `bright_radius` was removed in
                  release 4.0

        Returns
        -------
        qs : `ginga.misc.Bunch.Bunch`
            This is a single element of ``objlist`` as described in
            :meth:`evaluate_peaks`.

        """
        if x1 is None:
            x1 = 0
        if y1 is None:
            y1 = 0
        if x2 is None:
            x2 = image.width
        if y2 is None:
            y2 = image.height

        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        data = image.cutout_data(x1, y1, x2, y2, astype=float)

        qs = self.pick_field(data, peak_radius=radius,
                             fwhm_radius=fwhm_radius,
                             threshold=threshold,
                             minfwhm=minfwhm, maxfwhm=maxfwhm,
                             minelipse=minelipse, edgew=edgew,
                             ee_total_radius=ee_total_radius)

        # Add back in offsets into image to get correct values with respect
        # to the entire image
        qs.x += x1
        qs.y += y1
        qs.objx += x1
        qs.objy += y1
        self.logger.debug("obj=%f,%f fwhm=%f sky=%f bright=%f" % (
            qs.objx, qs.objy, qs.fwhm, qs.skylevel, qs.brightness))

        return qs

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.gw import Widgets
from ginga.util import rgb_cms
from ginga import trcalc

from .base import Stage, StageAction

icc_profiles = list(rgb_cms.get_profiles())
icc_profiles.insert(0, None)
icc_profiles.insert(0, 'working')
icc_intents = list(rgb_cms.get_intents())


class ICCProf(Stage):
    """Convert the given RGB data from the input ICC profile
    to the output ICC profile.
    """
    _stagename = 'icc-profile'

    def __init__(self):
        super().__init__()

        self._icc_input_profile = 'working'
        self._icc_output_profile = 'working'
        self._icc_output_intent = 'perceptual'
        self._icc_proof_profile = None
        self._icc_proof_intent = 'perceptual'
        self._icc_black_point_compensation = False

    def build_gui(self, container):

        captions = (('Input ICC profile:', 'label', 'Input ICC profile',
                     'combobox'),
                    ('Output ICC profile:', 'label', 'Output ICC profile',
                     'combobox'),
                    ('Rendering intent:', 'label', 'Rendering intent',
                     'combobox'),
                    ('Proof ICC profile:', 'label', 'Proof ICC profile',
                     'combobox'),
                    ('Proof intent:', 'label', 'Proof intent', 'combobox'),
                    ('__x', 'spacer', 'Black point compensation', 'checkbutton'),
                    )
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)

        value = self._icc_input_profile
        combobox = b.input_icc_profile
        for name in icc_profiles:
            combobox.append_text(str(name))
        try:
            index = icc_profiles.index(value)
            combobox.set_index(index)
        except Exception:
            pass
        combobox.add_callback('activated', self.set_icc_profile_cb)
        combobox.set_tooltip("ICC profile for the stage input")

        value = self._icc_output_profile
        combobox = b.output_icc_profile
        for name in icc_profiles:
            combobox.append_text(str(name))
        try:
            index = icc_profiles.index(value)
            combobox.set_index(index)
        except Exception:
            pass
        combobox.add_callback('activated', self.set_icc_profile_cb)
        combobox.set_tooltip("ICC profile for the stage output")

        value = self._icc_output_intent
        combobox = b.rendering_intent
        for name in icc_intents:
            combobox.append_text(name)
        try:
            index = icc_intents.index(value)
            combobox.set_index(index)
        except Exception:
            pass
        combobox.add_callback('activated', self.set_icc_profile_cb)
        combobox.set_tooltip("Rendering intent for the viewer display")

        value = self._icc_proof_profile
        combobox = b.proof_icc_profile
        for name in icc_profiles:
            combobox.append_text(str(name))
        try:
            index = icc_profiles.index(value)
            combobox.set_index(index)
        except Exception:
            pass
        combobox.add_callback('activated', self.set_icc_profile_cb)
        combobox.set_tooltip("ICC profile for soft proofing")

        value = self._icc_proof_intent
        combobox = b.proof_intent
        for name in icc_intents:
            combobox.append_text(name)
        try:
            index = icc_intents.index(value)
            combobox.set_index(index)
        except Exception:
            pass
        combobox.add_callback('activated', self.set_icc_profile_cb)
        combobox.set_tooltip("Rendering intent for soft proofing")

        value = self._icc_black_point_compensation
        b.black_point_compensation.set_state(value)
        b.black_point_compensation.add_callback(
            'activated', self.set_icc_profile_cb)
        b.black_point_compensation.set_tooltip("Use black point compensation")

        fr = Widgets.Frame()
        fr.set_widget(w)

        container.set_widget(fr)

    @property
    def icc_input_profile(self):
        return self._icc_input_profile

    @icc_input_profile.setter
    def icc_input_profile(self, val):
        self._icc_input_profile = val
        if self.gui_up:
            idx = icc_profiles.index(val)
            self.w.input_icc_profile.set_index(idx)

    @property
    def icc_output_profile(self):
        return self._icc_output_profile

    @icc_output_profile.setter
    def icc_output_profile(self, val):
        self._icc_output_profile = val
        if self.gui_up:
            idx = icc_profiles.index(val)
            self.w.output_icc_profile.set_index(idx)

    @property
    def icc_output_intent(self):
        return self._icc_output_intent

    @icc_output_intent.setter
    def icc_output_intent(self, val):
        self._icc_output_intent = val
        if self.gui_up:
            idx = icc_intents.index(val)
            self.w.rendering_intent.set_index(idx)

    @property
    def icc_proof_profile(self):
        return self._icc_proof_profile

    @icc_proof_profile.setter
    def icc_proof_profile(self, val):
        self._icc_proof_profile = val
        if self.gui_up:
            idx = icc_profiles.index(val)
            self.w.proof_icc_profile.set_index(idx)

    @property
    def icc_proof_intent(self):
        return self._icc_proof_intent

    @icc_proof_intent.setter
    def icc_proof_intent(self, val):
        self._icc_proof_intent = val
        if self.gui_up:
            idx = icc_intents.index(val)
            self.w.proof_intent.set_index(idx)

    @property
    def icc_black_point_compensation(self):
        return self._icc_black_point_compensation

    @icc_black_point_compensation.setter
    def icc_black_point_compensation(self, val):
        self._icc_black_point_compensation = val
        if self.gui_up:
            self.w.black_point_compensation.set_state(val)

    def set_icc_profile_cb(self, setting, idx):
        idx = self.w.input_icc_profile.get_index()
        input_profile_name = icc_profiles[idx]
        idx = self.w.output_icc_profile.get_index()
        output_profile_name = icc_profiles[idx]
        idx = self.w.rendering_intent.get_index()
        intent_name = icc_intents[idx]

        idx = self.w.proof_icc_profile.get_index()
        proof_profile_name = icc_profiles[idx]
        idx = self.w.proof_intent.get_index()
        proof_intent = icc_intents[idx]

        bpc = self.w.black_point_compensation.get_state()

        old = dict(icc_input_profile=self._icc_input_profile,
                   icc_output_profile=self._icc_output_profile,
                   icc_output_intent=self._icc_output_intent,
                   icc_proof_profile=self._icc_proof_profile,
                   icc_proof_intent=self._icc_proof_intent,
                   icc_black_point_compensation=self._icc_black_point_compensation)
        self._icc_input_profile = input_profile_name
        self._icc_output_profile = output_profile_name
        self._icc_output_intent = intent_name
        self._icc_proof_profile = proof_profile_name
        self._icc_proof_intent = proof_intent
        self._icc_black_point_compensation = bpc
        new = dict(icc_input_profile=self._icc_input_profile,
                   icc_output_profile=self._icc_output_profile,
                   icc_output_intent=self._icc_output_intent,
                   icc_proof_profile=self._icc_proof_profile,
                   icc_proof_intent=self._icc_proof_intent,
                   icc_black_point_compensation=self._icc_black_point_compensation)
        self.pipeline.push(StageAction(self, old, new,
                                       descr="iccprof / change"))

        self.pipeline.run_from(self)

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        input_profile = self._icc_input_profile
        if input_profile == 'working':
            input_profile = rgb_cms.working_profile
        # TODO
        #if input_profile is None:
        #    input_profile = 'sRGB'

        output_profile = self._icc_output_profile
        if output_profile == 'working':
            output_profile = rgb_cms.working_profile

        if (self._bypass or data is None or
            None in [input_profile, output_profile] or
            input_profile == output_profile):
            self.pipeline.set(icc_output_profile=output_profile)
            self.pipeline.send(res_np=data)
            return

        # color profiling will not work with other types, currently
        data = data.astype(np.uint8)

        alpha = None
        ht, wd, dp = data.shape
        if dp > 3:
            # color profile conversion does not handle an alpha layer
            alpha = data[:, :, 3]
            data = data[:, :, 0:3]

        try:
            arr = rgb_cms.convert_profile_fromto(data,
                                                 input_profile,
                                                 output_profile,
                                                 to_intent=self._icc_output_intent,
                                                 proof_name=self._icc_proof_profile,
                                                 proof_intent=self._icc_proof_intent,
                                                 use_black_pt=self._icc_black_point_compensation,
                                                 logger=self.logger)

            self.pipeline.set(icc_output_profile=output_profile)
            self.logger.debug("Converted from '%s' to '%s' profile" % (
                input_profile, output_profile))

        except Exception as e:
            self.logger.warning("Error converting between profiles: %s" % (str(e)))
            # TODO: maybe should have a traceback here
            self.logger.info("Stage output left unprofiled")
            self.pipeline.set(icc_output_profile=input_profile)
            arr = data

        if alpha is not None:
            arr = trcalc.add_alpha(arr, alpha)

        self.pipeline.send(res_np=arr)

    def _get_state(self):
        return dict(icc_input_profile=self._icc_input_profile,
                    icc_output_profile=self._icc_output_profile,
                    icc_output_intent=self._icc_output_intent,
                    icc_proof_profile=self._icc_proof_profile,
                    icc_proof_intent=self._icc_proof_intent,
                    icc_black_point_compensation=self._icc_black_point_compensation)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.icc_input_profile = d['icc_input_profile']
        self.icc_output_profile = d['icc_output_profile']
        self.icc_output_intent = d['icc_output_intent']
        self.icc_proof_profile = d['icc_proof_profile']
        self.icc_proof_intent = d['icc_proof_intent']
        self.icc_black_point_compensation = d['icc_black_point_compensation']



"""Test ap_regions.py"""

import numpy as np
import pytest
import logging

regions = pytest.importorskip('regions')

from astropy import units as u
from astropy.coordinates import SkyCoord

from ginga.util.ap_region import (astropy_region_to_ginga_canvas_object as r2g,
                                  ginga_canvas_object_to_astropy_region as g2r)
from ginga.canvas.CanvasObject import get_canvas_types


dc = get_canvas_types()


class Test_R2G(object):
    """Test conversions from astropy-regions to Ginga canvas types."""

    def test_point_pix1(self):
        r = regions.PointPixelRegion(center=regions.PixCoord(x=42, y=43),
                                     visual=regions.RegionVisual(symbol='x'))
        o = r2g(r)
        assert isinstance(o, dc.Point) and o.style == 'cross'
        assert np.all(np.isclose((o.x, o.y), (r.center.x, r.center.y)))

    def test_point_pix2(self):
        r = regions.PointPixelRegion(center=regions.PixCoord(x=42, y=43),
                                     visual=regions.RegionVisual(symbol='+'))
        o = r2g(r)
        assert isinstance(o, dc.Point) and o.style == 'plus'

    def test_point_pix3(self):
        r = regions.PointPixelRegion(center=regions.PixCoord(x=42, y=43),
                                     visual=regions.RegionVisual(symbol='*'))
        o = r2g(r)
        assert isinstance(o, dc.Point) and o.style == 'square' and \
            o.coord in ('data', None)

    def test_point_sky1(self):
        r = regions.PointSkyRegion(center=SkyCoord(12.0, 10.0, unit='deg',
                                                   frame='fk5'),
                                   visual=regions.RegionVisual(symbol='*'))
        o = r2g(r)
        assert isinstance(o, dc.Point) and o.style == 'square' and \
            o.coord == 'wcs'

    def test_text_pix1(self):
        r = regions.TextPixelRegion(center=regions.PixCoord(x=42, y=43),
                                    text='Foo',
                                    visual=regions.RegionVisual(textangle='45'))
        o = r2g(r)
        assert isinstance(o, dc.Text) and o.text == 'Foo' and o.coord in ('data', None)
        assert np.all(np.isclose((o.x, o.y, o.rot_deg),
                                 (r.center.x, r.center.y, float(r.visual['textangle']))))

    def test_text_sky1(self):
        r = regions.TextSkyRegion(center=SkyCoord(12.0, 10.0, unit='deg',
                                                  frame='fk5'),
                                  text='Bar',
                                  visual=regions.RegionVisual(textangle='-50'))
        o = r2g(r)
        assert isinstance(o, dc.Text) and o.text == 'Bar' and o.coord == 'wcs'
        assert np.all(np.isclose((o.x, o.y, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  float(r.visual['textangle']))))

    def test_line_pix1(self):
        r = regions.LinePixelRegion(start=regions.PixCoord(x=42, y=43),
                                    end=regions.PixCoord(x=42, y=43))
        o = r2g(r)
        assert isinstance(o, dc.Line) and o.coord in ('data', None)
        assert np.all(np.isclose((o.x1, o.y1, o.x2, o.y2),
                                 (r.start.x, r.start.y, r.end.x, r.end.y)))

    def test_line_sky1(self):
        r = regions.LineSkyRegion(start=SkyCoord(12.0, 10.0, unit='deg',
                                                 frame='fk5'),
                                  end=SkyCoord(12.5, 9.3, unit='deg',
                                               frame='fk5'))
        o = r2g(r)
        assert isinstance(o, dc.Line) and o.coord == 'wcs'
        assert np.all(np.isclose((o.x1, o.y1, o.x2, o.y2),
                                 (r.start.ra.deg, r.start.dec.deg,
                                  r.end.ra.deg, r.end.dec.deg)))

    def test_rectangle_pix1(self):
        r = regions.RectanglePixelRegion(center=regions.PixCoord(x=42, y=43),
                                         width=3, height=4,
                                         angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Box)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.x, r.center.y, r.width * 0.5, r.height * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_rectangle_sky1(self):
        r = regions.RectangleSkyRegion(center=SkyCoord(12.0, 10.0, unit='deg',
                                                       frame='fk5'),
                                       width=0.2 * u.deg, height=0.4 * u.deg,
                                       angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Box) and o.coord == 'wcs'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.width.value * 0.5,
                                  r.height.value * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_circle_pix1(self):
        r = regions.CirclePixelRegion(center=regions.PixCoord(x=42, y=43),
                                      radius=4.2)
        o = r2g(r)
        assert isinstance(o, dc.Circle)
        assert np.all(np.isclose((o.x, o.y, o.radius), (r.center.x, r.center.y, r.radius)))

    def test_circle_sky1(self):
        r = regions.CircleSkyRegion(center=SkyCoord(12.0, 10.0, unit='deg',
                                                    frame='fk5'),
                                    radius=0.5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Circle) and o.coord == 'wcs'
        assert np.all(np.isclose((o.x, o.y, o.radius),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.radius.value)))

        # round-trip test
        r2 = g2r(o)
        assert isinstance(r2, regions.CircleSkyRegion)
        assert np.all(np.isclose((r.center.ra.deg, r.center.dec.deg, r.radius.value),
                                 (r2.center.ra.deg, r2.center.dec.deg, r2.radius.value)))

    def test_ellipse_pix1(self):
        r = regions.EllipsePixelRegion(center=regions.PixCoord(x=42, y=43),
                                       height=4.2, width=4.2, angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Ellipse)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.x, r.center.y, r.width * 0.5, r.height * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_ellipse_sky1(self):
        r = regions.EllipseSkyRegion(center=SkyCoord(12.0, 10.0, unit='deg',
                                                     frame='fk5'),
                                     height=0.4 * u.deg, width=0.2 * u.deg,
                                     angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Ellipse) and o.coord == 'wcs'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.width.value * 0.5, r.height.value * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_polygon_pix1(self):
        r = regions.PolygonPixelRegion(vertices=regions.PixCoord(x=[1, 2, 2], y=[1, 1, 2]))
        o = r2g(r)
        assert isinstance(o, dc.Polygon)
        assert np.all(np.isclose(o.points, np.array(r.vertices.xy).T))

    def test_polygon_sky1(self):
        r = regions.PolygonSkyRegion(vertices=SkyCoord([1, 2, 2], [1, 1, 2],
                                                       unit='deg', frame='fk5'))
        o = r2g(r)
        assert isinstance(o, dc.Polygon) and o.coord == 'wcs'
        vertices = np.array((r.vertices.ra.deg, r.vertices.dec.deg)).T
        assert np.all(np.isclose(o.points, vertices))

    def test_circle_annulus_pix1(self):
        r = regions.CircleAnnulusPixelRegion(center=regions.PixCoord(x=42, y=43),
                                             inner_radius=4.2, outer_radius=5.2)
        o = r2g(r)
        assert isinstance(o, dc.Annulus)
        assert np.all(np.isclose((o.x, o.y, o.radius, o.radius + o.width),
                                 (r.center.x, r.center.y, r.inner_radius, r.outer_radius)))

    def test_circle_annulus_sky1(self):
        r = regions.CircleAnnulusSkyRegion(center=SkyCoord(12.0, 10.0, unit='deg',
                                                           frame='fk5'),
                                           inner_radius=0.2 * u.deg,
                                           outer_radius=0.22 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Annulus) and o.coord == 'wcs'
        assert np.all(np.isclose((o.x, o.y, o.radius, o.radius + o.width),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.inner_radius.value, r.outer_radius.value)))

    def test_ellipse_annulus_pix1(self):
        r = regions.EllipseAnnulusPixelRegion(center=regions.PixCoord(x=42, y=43),
                                              inner_width=4.2, outer_width=5.2,
                                              inner_height=7.2, outer_height=8.2,
                                              angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Annulus2R) and o.atype == 'ellipse'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius,
                                  o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y,
                                  r.inner_width * 0.5, r.inner_height * 0.5,
                                  r.outer_width * 0.5, r.outer_height * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_ellipse_annulus_sky1(self):
        r = regions.EllipseAnnulusSkyRegion(center=SkyCoord(12.0, 10.0,
                                                            unit='deg',
                                                            frame='fk5'),
                                            inner_width=0.2 * u.deg,
                                            outer_width=0.4 * u.deg,
                                            inner_height=0.5 * u.deg,
                                            outer_height=0.6 * u.deg,
                                            angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Annulus2R) and o.atype == 'ellipse' \
            and o.coord == 'wcs'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius,
                                  o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.inner_width.value * 0.5,
                                  r.inner_height.value * 0.5,
                                  r.outer_width.value * 0.5,
                                  r.outer_height.value * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_rectangle_annulus_pix1(self):
        r = regions.RectangleAnnulusPixelRegion(center=regions.PixCoord(x=42, y=43),
                                                inner_width=4.2,
                                                outer_width=5.2,
                                                inner_height=7.2,
                                                outer_height=8.2,
                                                angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Annulus2R) and o.atype == 'box'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y, r.inner_width * 0.5, r.inner_height * 0.5,
                                  r.outer_width * 0.5, r.outer_height * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_rectangle_annulus_sky1(self):
        r = regions.RectangleAnnulusSkyRegion(center=SkyCoord(12.0, 10.0,
                                                              unit='deg',
                                                              frame='fk5'),
                                              inner_width=0.2 * u.deg,
                                              outer_width=0.4 * u.deg,
                                              inner_height=0.5 * u.deg,
                                              outer_height=0.6 * u.deg,
                                              angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Annulus2R) and o.atype == 'box' and o.coord == 'wcs'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.inner_width.value * 0.5,
                                  r.inner_height.value * 0.5,
                                  r.outer_width.value * 0.5,
                                  r.outer_height.value * 0.5,
                                  r.angle.to(u.deg).value)))


class Test_G2R(object):
    """Test conversions from Ginga canvas types to astropy-regions Regions."""

    def test_point_pix1(self):
        o = dc.Point(42, 43, radius=5, style='cross')
        r = g2r(o)
        assert isinstance(r, regions.PointPixelRegion)
        assert np.all(np.isclose((o.x, o.y), (r.center.x, r.center.y)))
        assert r.visual['symbol'] == 'x'

    def test_point_pix2(self):
        o = dc.Point(42, 43, radius=5, style='plus')
        r = g2r(o)
        assert isinstance(r, regions.PointPixelRegion)
        assert r.visual['symbol'] == '+'

    def test_point_pix3(self):
        o = dc.Point(42, 43, radius=5, style='square')
        r = g2r(o)
        assert isinstance(r, regions.PointPixelRegion)
        assert r.visual['symbol'] == '*'

    def test_point_sky1(self):
        o = dc.Point(12.0, 10.0, radius=0.001, style='plus', coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.PointSkyRegion)
        assert r.visual['symbol'] == '+'
        assert np.all(np.isclose((o.x, o.y),
                                 (r.center.ra.deg, r.center.dec.deg)))

    def test_text_pix1(self):
        o = dc.Text(42, 43, text='Foo', rot_deg=45.0)
        r = g2r(o)
        assert isinstance(r, regions.TextPixelRegion) and r.text == 'Foo'
        assert np.all(np.isclose((o.x, o.y, o.rot_deg),
                                 (r.center.x, r.center.y, float(r.visual['textangle']))))

    def test_text_sky1(self):
        o = dc.Text(12.0, 10.0, text='Bar', rot_deg=45.0, coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.TextSkyRegion) and r.text == 'Bar'
        assert np.all(np.isclose((o.x, o.y, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  float(r.visual['textangle']))))

    def test_line_pix1(self):
        o = dc.Line(42, 43, 52, 53, arrow='end', color='red')
        r = g2r(o)
        assert isinstance(r, regions.LinePixelRegion)
        assert r.visual['color'] == 'red'
        # lines ends (arrows) not supported yet, but from documentation
        # looks like they could be in the future?
        #assert r.visual['line'] == '[0 1]'
        assert np.all(np.isclose((o.x1, o.y1, o.x2, o.y2),
                                 (r.start.x, r.start.y, r.end.x, r.end.y)))

    def test_line_sky1(self):
        o = dc.Line(12.0, 10.0, 12.5, 9.3, color='blue', linewidth=2,
                    linestyle='dash', coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.LineSkyRegion)
        assert r.visual['color'] == 'blue'
        assert r.visual['linewidth'] == 2
        # ditto: dashed line
        #assert r.visual['dash'] == 1
        assert np.all(np.isclose((o.x1, o.y1, o.x2, o.y2),
                                 (r.start.ra.deg, r.start.dec.deg,
                                  r.end.ra.deg, r.end.dec.deg)))

    def test_box_pix1(self):
        o = dc.Box(42, 43, 10, 10)
        r = g2r(o)
        assert isinstance(r, regions.RectanglePixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius),
                                 (r.center.x, r.center.y, r.width * 0.5, r.height * 0.5)))

    def test_box_sky1(self):
        o = dc.Box(12.0, 10.0, 0.1, 0.2,
                   color='green', fill=True, fillcolor='orange', coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.RectangleSkyRegion)
        assert r.visual['color'] == 'green'
        assert r.visual['fill'] == 1
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.width.value * 0.5, r.height.value * 0.5)))

    def test_circle_pix1(self):
        o = dc.Circle(42, 43, 4.2)
        r = g2r(o)
        assert isinstance(r, regions.CirclePixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.radius), (r.center.x, r.center.y, r.radius)))

    def test_circle_sky1(self):
        o = dc.Circle(12.0, 10.0, 1.1, coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.CircleSkyRegion)
        assert np.all(np.isclose((o.x, o.y, o.radius),
                                 (r.center.ra.deg, r.center.dec.deg, r.radius.value)))

    def test_ellipse_pix1(self):
        o = dc.Ellipse(42, 43, 4.2, 4.2, rot_deg=5.0)
        r = g2r(o)
        assert isinstance(r, regions.EllipsePixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.x, r.center.y, r.width * 0.5, r.height * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_ellipse_sky1(self):
        o = dc.Ellipse(12.0, 10.0, 0.2, 0.4, rot_deg=5.0, coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.EllipseSkyRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.width.value * 0.5, r.height.value * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_polygon_pix1(self):
        o = dc.Polygon([(1, 1), (2, 1), (2, 2)], color='red', fill=True,
                       linewidth=2)
        r = g2r(o)
        assert isinstance(r, regions.PolygonPixelRegion)
        assert np.all(np.isclose(o.points, np.array(r.vertices.xy).T))

        # round-trip test
        o2 = r2g(r)
        assert isinstance(o2, dc.Polygon) and o2.coord in ('data', None)
        assert o2.color == o2.color and o2.linewidth == o.linewidth and \
            o2.fill == o.fill
        assert np.all(np.isclose(o.points, o2.points))

    def test_polygon_sky1(self):
        o = dc.Polygon([(1, 1), (2, 1), (2, 2)], coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.PolygonSkyRegion)
        vertices = np.array((r.vertices.ra.deg, r.vertices.dec.deg)).T
        assert np.all(np.isclose(o.points, vertices))

    def test_circle_annulus_pix1(self):
        o = dc.Annulus(42, 43, 4.2, width=1, atype='circle')
        r = g2r(o)
        assert isinstance(r, regions.CircleAnnulusPixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.radius, o.radius + o.width),
                                 (r.center.x, r.center.y, r.inner_radius, r.outer_radius)))

    def test_circle_annulus_sky1(self):
        o = dc.Annulus(12.0, 10.0, 1.0, width=0.01, atype='circle', coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.CircleAnnulusSkyRegion)
        assert np.all(np.isclose((o.x, o.y, o.radius, o.radius + o.width),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.inner_radius.value, r.outer_radius.value)))

    def test_ellipse_annulus_pix1(self):
        o = dc.Annulus2R(42, 43, 4.2, 7.2, xwidth=1, ywidth=1, rot_deg=5.0, atype='ellipse')
        r = g2r(o)
        assert isinstance(r, regions.EllipseAnnulusPixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y, r.inner_width * 0.5, r.inner_height * 0.5,
                                  r.outer_width * 0.5, r.outer_height * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_ellipse_annulus_sky1(self):
        o = dc.Annulus2R(12.0, 10.0, 1.0, 1.1, xwidth=0.1, ywidth=0.2,
                         rot_deg=5.0, atype='ellipse', coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.EllipseAnnulusSkyRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.inner_width.value * 0.5,
                                  r.inner_height.value * 0.5,
                                  r.outer_width.value * 0.5,
                                  r.outer_height.value * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_rectangle_annulus_pix1(self):
        o = dc.Annulus2R(42, 43, 4.2, 7.2, xwidth=1, ywidth=1, rot_deg=5.0, atype='box')
        r = g2r(o)
        assert isinstance(r, regions.RectangleAnnulusPixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y, r.inner_width * 0.5, r.inner_height * 0.5,
                                  r.outer_width * 0.5, r.outer_height * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_rectangle_annulus_sky1(self):
        o = dc.Annulus2R(12.0, 10.0, 1.0, 1.1, xwidth=0.1, ywidth=0.2,
                         rot_deg=5.0, atype='box', coord='wcs')
        r = g2r(o)
        assert isinstance(r, regions.RectangleAnnulusSkyRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.ra.deg, r.center.dec.deg,
                                  r.inner_width.value * 0.5,
                                  r.inner_height.value * 0.5,
                                  r.outer_width.value * 0.5,
                                  r.outer_height.value * 0.5,
                                  r.angle.to(u.deg).value)))

    def test_skip_error(self):
        logger = logging.getLogger("test_ap_regions")
        o = dc.Crosshair(0, 0)
        r = g2r(o, logger=logger)
        assert isinstance(r, regions.TextPixelRegion)
        assert np.all(np.isclose((o.x, o.y), (r.center.x, r.center.y)))



"""Tools for accessing HTML doc from RTD."""

import re

from ginga.GingaPlugin import GlobalPlugin, LocalPlugin
import ginga

__all__ = ['get_online_docs_url']

# base of our online documentation
rtd_base_url = "https://ginga.readthedocs.io/en/"


def get_online_docs_url(plugin=None):
    """
    Return URL to online documentation closest to this Ginga version.

    Parameters
    ----------
    plugin : obj or `None`
        Plugin object. If given, URL points to plugin doc directly.
        If this function is called from within plugin class,
        pass ``self`` here.

    Returns
    -------
    url : str
        URL to online documentation (top-level, if plugin == None).

    """
    ginga_ver = ginga.__version__
    if re.match(r'^v\d+\.\d+\.\d+$', ginga_ver):
        rtd_version = ginga_ver
    else:
        # default to latest
        rtd_version = 'latest'
    url = f"{rtd_base_url}{rtd_version}"
    if plugin is not None:
        plugin_name = str(plugin)
        if isinstance(plugin, GlobalPlugin):
            url += f'/manual/plugins_global/{plugin_name}.html'
        elif isinstance(plugin, LocalPlugin):
            url += f'/manual/plugins_local/{plugin_name}.html'

    return url



#
# camera.py -- mode for operating OpenGL camera
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""Camera Mode enables bindings that can manipulate the camera in
a Ginga image viewer operating with a OpenGL backend.

.. note:: Camera Mode does not work unless the viewer is using an
          OpenGL backend.

Enter the mode by
-----------------
* Space, then "c"

Exit the mode by
----------------
* Esc

Default bindings in mode
------------------------
* s : save the current camera settings, to restore later
* r : reset to the saved camera settings
* 3 : change the camera view into 3D mode
* scroll : track camera (dolly in and out)
* left drag : camera orbit the view
* right drag : camera pan the view

"""
from ginga.modes.mode_base import Mode


class CameraMode(Mode):

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_camera=['__c', None, 'pan'],

            kp_camera_save=['camera+s'],
            kp_camera_reset=['camera+r'],
            kp_camera_toggle3d=['camera+3'],

            sc_camera_track=['camera+scroll'],

            ms_camera_orbit=['camera+left'],
            ms_camera_pan_delta=['camera+right'])

    def __str__(self):
        return 'camera'

    def start(self):
        pass

    def stop(self):
        self.onscreen_message(None)

    def get_camera(self, viewer):
        renderer = viewer.renderer
        if not hasattr(renderer, 'camera'):
            return None, None
        return renderer.camera

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_camera_reset(self, viewer, event, data_x, data_y):
        event.accept()
        camera = self.get_camera(viewer)
        if camera is None:
            # this viewer doesn't have a camera
            return False

        camera.reset()
        camera.calc_gl_transform()
        self.onscreen_message("Reset camera", delay=0.5)
        viewer.update_widget()

    def kp_camera_save(self, viewer, event, data_x, data_y):
        event.accept()
        camera = self.get_camera(viewer)
        if camera is None:
            # this viewer doesn't have a camera
            return False

        camera.save_positions()
        self.onscreen_message("Saved camera position", delay=0.5)

    def kp_camera_toggle3d(self, viewer, event, data_x, data_y):
        event.accept()
        camera = self.get_camera(viewer)
        if camera is None:
            # this viewer doesn't have a camera
            return False

        renderer = viewer.renderer
        renderer.mode3d = not renderer.mode3d
        viewer.update_widget()

    #####  SCROLL ACTION CALLBACKS #####

    def sc_camera_track(self, viewer, event, msg=True):
        camera = self.get_camera(viewer)
        if camera is None:
            # this viewer doesn't have a camera
            return False
        event.accept()

        zoom_accel = self.settings.get('scroll_zoom_acceleration', 6.0)
        delta = event.amount * zoom_accel

        direction = self.get_direction(event.direction)
        if direction == 'down':
            delta = - delta

        camera.track(delta)
        camera.calc_gl_transform()

        scales = camera.get_scale_2d()
        # TODO: need to set scale in viewer settings, without triggering a
        # scale operation on this viewer
        viewer.update_widget()

    #####  MOUSE ACTION CALLBACKS #####

    def ms_camera_orbit(self, viewer, event, data_x, data_y, msg=True):
        camera = self.get_camera(viewer)
        if camera is None:
            # this viewer doesn't have a camera
            return False
        event.accept()

        x, y = self.get_win_xy(viewer)

        if event.state == 'move':
            camera.orbit(self._start_x, self._start_y, x, y)
            self._start_x, self._start_y = x, y
            camera.calc_gl_transform()
            ## pos = tuple(camera.position.get())
            ## mst = "Camera position: (%.4f, %.4f, %.4f)" % pos
            ## if msg:
            ##     self.onscreen_message(mst, delay=0.5)
            tup = camera.position.get()

        elif event.state == 'down':
            self._start_x, self._start_y = x, y

        ## else:
        ##     self.onscreen_message(None)

        viewer.update_widget()

    def ms_camera_pan_delta(self, viewer, event, data_x, data_y, msg=True):
        camera = self.get_camera(viewer)
        if camera is None:
            # this viewer doesn't have a camera
            return False
        event.accept()

        x, y = self.get_win_xy(viewer)

        if event.state == 'move':
            dx, dy = x - self._start_x, self._start_y - y
            camera.pan_delta(dx, dy)
            self._start_x, self._start_y = x, y
            camera.calc_gl_transform()

        elif event.state == 'down':
            self._start_x, self._start_y = x, y
            ## if msg:
            ##     self.onscreen_message("Camera translate", delay=1.0)

        ## else:
        ##     self.onscreen_message(None)

        # TODO: need to get the updated pan position and set it in
        # viewer's settings without triggering a callback to the viewer
        # itself
        tup = camera.position.get()
        data_x, data_y = viewer.tform['data_to_native'].from_(tup[:2])

        viewer.update_widget()



#
# RGBMap.py -- color mapping
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import uuid
import numpy as np

from ginga.misc import Callback, Settings, Bunch
from ginga import ColorDist, trcalc
from ginga import cmap as mod_cmap
from ginga import imap as mod_imap
from ginga.util import pipeline
from ginga.util.stages.color import (RGBInput, Distribute, ShiftMap,
                                     IntensityMap, ColorMap, RGBMapError)


class RGBPlanes(object):

    def __init__(self, rgbarr, order):
        self.rgbarr = rgbarr
        order = order.upper()
        self.order = order
        self.hasAlpha = 'A' in order

    def get_slice(self, ch):
        return self.rgbarr[..., self.order.index(ch.upper())]

    def has_slice(self, ch):
        return ch.upper() in self.order

    def get_order(self):
        return self.order

    def get_order_indexes(self, cs):
        cs = cs.upper()
        return [self.order.index(c) for c in cs]

    def get_array(self, order, dtype=None):
        """Get Numpy array that represents the RGB layers.

        Parameters
        ----------
        order : str
            The desired order of RGB color layers.

        Returns
        -------
        arr : ndarray
            Numpy array that represents the RGB layers.

        dtype : numpy array dtype (optional)
            Type of the numpy array desired.

        """
        if dtype is None:
            dtype = self.rgbarr.dtype
        order = order.upper()
        if order == self.order:
            return self.rgbarr.astype(dtype, copy=False)

        res = trcalc.reorder_image(order, self.rgbarr, self.order)
        res = res.astype(dtype, copy=False, casting='unsafe')

        return res

    def get_size(self):
        """Returns (height, width) tuple of slice size."""
        return self.get_slice('R').shape


class RGBMapper(Callback.Callbacks):

    ############################################################
    # CODE NOTES
    #
    # [A] Some numpy routines have been optomized by using the out=
    # parameter, which avoids having to allocate a new array for the
    # result
    #

    def __init__(self, logger, dist=None, settings=None, bpp=None):
        Callback.Callbacks.__init__(self)

        self.logger = logger
        self.mapper_id = str(uuid.uuid4())
        self.cache_arr = None

        # For color and intensity maps
        self.cmap = None
        self.imap = None
        self.dist = dist

        # targeted bit depth per-pixel band of the output RGB array
        # (can be less than the data size of the output array)
        if bpp is None:
            bpp = 8
        self.set_bpp(bpp)

        # Create settings and set defaults
        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings
        self.t_ = settings
        self.settings_keys = ['color_map', 'intensity_map',
                              'color_algorithm', 'color_hashsize',
                              'color_map_invert', 'color_map_rot_pct',
                              'contrast', 'brightness',
                              ]

        # add our defaults
        self.t_.add_defaults(color_map='gray', intensity_map='ramp',
                             color_map_invert=False, color_map_rot_pct=0.0,
                             color_algorithm='linear',
                             color_hashsize=self.maxc + 1,
                             contrast=0.5, brightness=0.5)
        self.t_.get_setting('color_map').add_callback('set',
                                                      self.color_map_set_cb)
        self.t_.get_setting('intensity_map').add_callback('set',
                                                          self.intensity_map_set_cb)
        self.t_.get_setting('color_map_invert').add_callback('set',
                                                             self.color_map_invert_set_cb)
        self.t_.get_setting('color_map_rot_pct').add_callback('set',
                                                              self.color_map_rot_pct_set_cb)
        self.t_.get_setting('contrast').add_callback('set',
                                                     self.contrast_set_cb)
        self.t_.get_setting('brightness').add_callback('set',
                                                       self.brightness_set_cb)
        self.t_.get_setting('color_hashsize').add_callback('set',
                                                           self.color_hashsize_set_cb)
        self.t_.get_setting('color_algorithm').add_callback('set',
                                                            self.color_algorithm_set_cb)

        # For callbacks
        for name in ('changed', ):
            self.enable_callback(name)

        self.suppress_changed = self.suppress_callback('changed', 'last')

        cm_name = self.t_.get('color_map', 'gray')
        self.set_color_map(cm_name)

        self.p_cmap.invert_cmap(False)
        self.p_cmap.rotate_color_map(0.0)

        im_name = self.t_.get('intensity_map', 'ramp')
        self.set_intensity_map(im_name)

    def set_bpp(self, bpp):
        """
        Set the bit depth (per band) for the output.
        A typical "32-bit RGBA" image would be 8; a 48-bit image would
        be 16, etc.
        """
        self.bpp = bpp
        self.maxc = int(2 ** self.bpp - 1)

        self.create_pipeline()

        self._refresh_cache()

    def create_pipeline(self):
        # create RGB mapping pipeline
        self.p_input = RGBInput(bpp=self.bpp)
        self.p_dist = Distribute(bpp=self.bpp)
        if self.dist is not None:
            self.p_dist.set_dist(self.dist)
        self.p_shift = ShiftMap(bpp=self.bpp)
        self.p_imap = IntensityMap(bpp=self.bpp)
        self.p_cmap = ColorMap(bpp=self.bpp)

        stages = [self.p_input,
                  self.p_dist,
                  self.p_shift,
                  self.p_imap,
                  self.p_cmap,
                  ]
        self.pipeline = pipeline.Pipeline(self.logger, stages)
        self.pipeline.name = 'rgb-mapper'

        # will be set from outer pipelines
        state = Bunch.Bunch(order='RGBA')
        self.pipeline.set(state=state)

        self._refresh_cache()

    def get_settings(self):
        return self.t_

    def set_color_map(self, cmap_name):
        self.t_.set(color_map=cmap_name)

    def color_map_set_cb(self, setting, cmap_name):
        cm = mod_cmap.get_cmap(cmap_name)
        self.set_cmap(cm, callback=True)

    def set_cmap(self, cmap, callback=True):
        """
        Set the color map used by this RGBMapper.

        `cmap` specifies a ColorMap object.  If `callback` is True, then
        any callbacks associated with this change will be invoked.
        """
        self.cmap = cmap
        with self.suppress_changed:
            self.calc_cmap()
            # TEMP: ignore passed callback parameter
            # callback=False in the following because we don't want to
            # recursively invoke set_cmap()
            self.t_.set(color_map=cmap.name, callback=False)

    def get_cmap(self):
        """
        Return the color map used by this RGBMapper.
        """
        return self.cmap

    def calc_cmap(self):
        self.p_cmap.set_cmap(self.cmap)
        self.recalc()

    def invert_cmap(self, callback=True):
        tf = not self.t_.get('color_map_invert', False)
        self.t_.set(color_map_invert=tf, callback=callback)

    def color_map_invert_set_cb(self, setting, tf):
        self.p_cmap.invert_cmap(tf)
        self.recalc()

    def rotate_cmap(self, num, callback=True):
        pct = num / (self.maxc + 1)
        self.t_.set(color_map_rot_pct=pct, callback=callback)

    def color_map_rot_pct_set_cb(self, setting, pct):
        self.p_cmap.rotate_color_map(pct)
        self.recalc()

    def restore_cmap(self, callback=True):
        """Undoes color map rotation and inversion, also resets contrast
        and brightness.
        """
        with self.suppress_changed:
            self.t_.set(color_map_invert=False, color_map_rot_pct=0.0,
                        contrast=0.5, brightness=0.5)

    def reset_cmap(self):
        """Similar to restore_cmap, but also restores the default color
        and intensity maps.
        """
        with self.suppress_changed:
            self.t_.set(color_map='gray', intensity_map='ramp')
            self.restore_cmap()

    def get_rgb(self, index):
        """
        Return a tuple of (R, G, B) values in the 0-maxc range associated
        mapped by the value of `index`.
        """
        index = int(index)
        assert (index >= 0) and (index <= self.maxc), \
            RGBMapError("Index must be in range 0-%d !" % (self.maxc))
        return tuple(self.cache_arr[index])

    def get_rgbval(self, index):
        """
        Return a tuple of (R, G, B) values in the 0-maxc range associated
        mapped by the value of `index`.
        """
        index = int(index)
        assert (index >= 0) and (index <= self.maxc), \
            RGBMapError("Index must be in range 0-%d !" % (self.maxc))
        return self.cache_arr[index]

    def get_colors(self):
        return self.cache_arr

    def set_intensity_map(self, imap_name):
        self.t_.set(intensity_map=imap_name)

    def intensity_map_set_cb(self, setting, imap_name):
        im = mod_imap.get_imap(imap_name)
        self.set_imap(im, callback=True)

    def set_imap(self, imap, callback=True):
        """
        Set the intensity map used by this RGBMapper.

        `imap` specifies an IntensityMap object.  If `callback` is True, then
        any callbacks associated with this change will be invoked.
        """
        self.imap = imap
        self.calc_imap()
        with self.suppress_changed:
            # TEMP: ignore passed callback parameter
            self.recalc(callback=callback)
            # callback=False in the following because we don't want to
            # recursively invoke set_imap()
            self.t_.set(intensity_map=imap.name, callback=False)

    def get_imap(self):
        """
        Return the intensity map used by this RGBMapper.
        """
        return self.imap

    def calc_imap(self):
        self.p_imap.set_imap(self.imap)
        self.recalc()

    def reset_sarr(self, callback=True):
        self.t_.set(contrast=0.5, brightness=0.5, callback=callback)

    def contrast_set_cb(self, setting, pct):
        self.p_shift.set_contrast(pct)
        self.recalc()

    def brightness_set_cb(self, setting, pct):
        self.p_shift.set_brightness(pct)
        self.recalc()

    def _refresh_cache(self):
        i_arr = np.arange(0, self.maxc + 1, dtype=np.uint)
        self.p_dist.result.setvals(res_np=i_arr)
        self.pipeline.run_from(self.p_shift)
        cache_arr = self.pipeline.get_data(self.pipeline[-1])
        if cache_arr is not None:
            if len(cache_arr.shape) < 2:
                # if colormap stage is bypassed then output will need to
                # be broadcast to RGB
                cache_arr = np.dstack((cache_arr, cache_arr, cache_arr)).reshape(cache_arr.shape + (3,))
            self.cache_arr = cache_arr[:, :3]

    def recalc(self, callback=True):
        self._refresh_cache()
        if callback:
            self.make_callback('changed')

    def get_hash_size(self):
        return self.p_dist.get_hash_size()

    def set_hash_size(self, size, callback=True):
        # TEMP: ignore passed callback parameter
        self.t_.set(color_hashsize=size)

    def color_hashsize_set_cb(self, setting, size):
        self.p_dist.set_hash_size(size)
        self.make_callback('changed')

    def get_hash_algorithms(self):
        return ColorDist.get_dist_names()

    def get_hash_algorithm(self):
        return str(self.p_dist.dist)

    def get_dist(self):
        """
        Return the color distribution used by this RGBMapper.
        """
        return self.p_dist.get_dist()

    def set_dist(self, dist, callback=True):
        self.p_dist.set_dist(dist)
        if callback:
            self.make_callback('changed')

    def set_hash_algorithm(self, name, callback=True, **kwargs):
        # TODO: deal with algorithm parameters in kwargs
        self.t_.set(color_algorithm=name)

    def set_color_algorithm(self, name):
        self.t_.set(color_algorithm=name)

    def color_algorithm_set_cb(self, setting, name):
        self.p_dist.set_color_algorithm(name)
        self.make_callback('changed')

    def get_order_indexes(self, order, cs):
        order = order.upper()
        if order == '':
            # assume standard RGB order if we don't find an order
            # explicitly set
            return [0, 1, 2]
        cs = cs.upper()
        return [order.index(c) for c in cs]

    def get_rgb_array(self, idx, order=None):

        if self.cache_arr is not None:
            # if the cache array is set, then this will deliver a faster
            # short cut to the colorized output--just apply the color
            # distribution and then map through the cache array
            idx = self.p_dist.get_hasharray(idx)
            arr_out = self.p_cmap.do_map_index(idx, self.cache_arr)

        else:
            # else run through the pipeline as usual
            self.p_input.set_input(idx)
            self.pipeline.run_from(self.p_input)

            arr_out = self.pipeline.get_data(self.pipeline[-1])

        # reorder as caller needs it
        state = self.pipeline.get('state')
        if order is not None and order != state.order:
            arr_out = trcalc.reorder_image(order, arr_out, state.order)

        return arr_out

    def get_hasharray(self, idx):
        return self.p_dist.dist.hash_array(idx)

    def copy_attributes(self, dst_rgbmap, keylist=None):
        if keylist is None:
            keylist = self.settings_keys
        self.t_.copy_settings(dst_rgbmap.get_settings(), keylist=keylist)

    def share_attributes(self, dst_rgbmap, keylist=None):
        if keylist is None:
            keylist = self.settings_keys
        self.t_.share_settings(dst_rgbmap.get_settings(), keylist=keylist)


class NonColorMapper(RGBMapper):
    """
    A kind of color mapper for data that is already in RGB form.

    This mapper allows changing of color distribution and contrast
    adjustment, but does no coloring.
    """
    def __init__(self, logger, dist=None, bpp=None):
        super().__init__(logger, dist=dist, bpp=bpp)

        self.p_imap.bypass(True)
        self.p_cmap.bypass(True)


class PassThruRGBMapper(RGBMapper):
    """
    A kind of color mapper for data that is already in RGB form.

    This mapper bypasses color distribution, contrast adjustment and
    coloring.  It is thus the most efficient one to use for maximum
    speed rendering of "finished" RGB data.
    """
    def __init__(self, logger, dist=None, bpp=None):
        super().__init__(logger, dist=dist, bpp=bpp)

        self.p_dist.bypass(True)
        self.p_shift.bypass(True)
        self.p_imap.bypass(True)
        self.p_cmap.bypass(True)



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga import ColorDist
from ginga import cmap as mod_cmap
from ginga import imap as mod_imap
from ginga.RGBMap import RGBMapper
from ginga.gw import Widgets, ColorBar

from .base import Stage, StageAction


class RGBMap(Stage):

    _stagename = 'rgb-mapper'

    def __init__(self):
        super().__init__()

        self.rgbmap = None
        self.calg_names = ColorDist.get_dist_names()
        self.cmap_names = mod_cmap.get_names()
        self.imap_names = mod_imap.get_names()
        self.order = 'RGB'

        self._calg_name = 'linear'
        self._cmap_name = 'gray'
        self._imap_name = 'ramp'
        self._invert_cmap = False
        self._rotate_cmap_pct = 0.0
        self._contrast = 0.5
        self._brightness = 0.5

        self.viewer = None
        self.fv = None

    def build_gui(self, container):
        self.viewer = self.pipeline.get('viewer')
        self.fv = self.pipeline.get("fv")
        top = Widgets.VBox()

        # create and initialize RGB mapper
        # TODO: currently hash size cannot be changed
        self.rgbmap = RGBMapper(self.logger)
        self.rgbmap.get_settings().set(color_algorithm=self._calg_name,
                                       color_map=self._cmap_name,
                                       intensity_map=self._imap_name,
                                       color_map_invert=self._invert_cmap,
                                       color_map_rot_pct=self._rotate_cmap_pct,
                                       contrast=self._contrast,
                                       brightness=self._brightness)
        self.rgbmap.add_callback('changed', self.rgbmap_changed_cb)

        self.settings_keys = list(self.rgbmap.settings_keys)
        self.settings_keys.remove('color_hashsize')

        fr = Widgets.Frame("Color Distribution")

        captions = (('Algorithm:', 'label', 'Algorithm', 'combobox'),
                    #('Table Size:', 'label', 'Table Size', 'entryset'),
                    ('Dist Defaults', 'button'))

        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)
        self.w.calg_choice = b.algorithm
        #self.w.table_size = b.table_size
        b.algorithm.set_tooltip("Choose a color distribution algorithm")
        #b.table_size.set_tooltip("Set size of the distribution hash table")
        b.dist_defaults.set_tooltip("Restore color distribution defaults")
        b.dist_defaults.add_callback('activated',
                                     lambda w: self.set_default_distmaps())

        combobox = b.algorithm
        options = []
        for name in self.calg_names:
            options.append(name)
            combobox.append_text(name)
        try:
            index = self.calg_names.index(self._calg_name)
            combobox.set_index(index)
        except Exception:
            pass
        combobox.add_callback('activated', self.set_calg_cb)

        fr.set_widget(w)
        top.add_widget(fr, stretch=0)

        # COLOR MAPPING OPTIONS
        fr = Widgets.Frame("Color Mapping")

        captions = (('Colormap:', 'label', 'Colormap', 'combobox'),
                    ('Intensity:', 'label', 'Intensity', 'combobox'),
                    ('Rotate:', 'label', 'rotate_cmap', 'hscale'),
                    ('Invert CMap', 'checkbutton', 'Unrotate CMap', 'button'),
                    ('Color Defaults', 'button', 'Copy from viewer', 'button'))
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)
        self.w.cmap_choice = b.colormap
        self.w.imap_choice = b.intensity

        b.invert_cmap.set_tooltip("Invert color map")
        b.invert_cmap.set_state(False)
        b.invert_cmap.add_callback('activated', self.invert_cmap_cb)

        b.rotate_cmap.set_tracking(False)
        b.rotate_cmap.set_limits(0, 100, incr_value=1)
        b.rotate_cmap.set_value(0)
        b.rotate_cmap.add_callback('value-changed', self.rotate_cmap_cb)
        b.rotate_cmap.set_tooltip("Rotate the colormap")

        b.colormap.set_tooltip("Choose a color map for this image")
        b.intensity.set_tooltip("Choose an intensity map for this image")
        b.unrotate_cmap.set_tooltip("Undo cmap rotation")
        b.unrotate_cmap.add_callback('activated', self.unrotate_cmap_cb)
        b.color_defaults.set_tooltip("Restore all color map settings to defaults")
        b.color_defaults.add_callback('activated',
                                      lambda w: self.set_default_cmaps())
        fr.set_widget(w)

        combobox = b.colormap
        options = []
        for name in self.cmap_names:
            options.append(name)
            combobox.append_text(name)
        try:
            index = self.cmap_names.index(self._cmap_name)
        except Exception:
            index = self.cmap_names.index('gray')
        combobox.set_index(index)
        combobox.add_callback('activated', self.set_cmap_cb)

        combobox = b.intensity
        options = []
        for name in self.imap_names:
            options.append(name)
            combobox.append_text(name)
        try:
            index = self.imap_names.index(self._imap_name)
        except Exception:
            index = self.imap_names.index('ramp')
        combobox.set_index(index)
        combobox.add_callback('activated', self.set_imap_cb)

        b.copy_from_viewer.set_tooltip("Copy settings from viewer")
        b.copy_from_viewer.add_callback('activated', self.copy_from_viewer_cb)

        top.add_widget(fr, stretch=0)

        # CONTRAST MANIPULATIONS
        fr = Widgets.Frame("Contrast and Brightness (Bias)")

        captions = (('Contrast:', 'label', 'contrast', 'hscale'),
                    ('Brightness:', 'label', 'brightness', 'hscale'),
                    ('_cb1', 'spacer', '_hbox_cb', 'hbox'))
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)

        b.contrast.set_tracking(False)
        b.contrast.set_limits(0, 100, incr_value=1)
        b.contrast.set_value(50)
        b.contrast.add_callback('value-changed', self.contrast_set_cb)
        b.contrast.set_tooltip("Set contrast for the viewer")

        b.brightness.set_tracking(False)
        b.brightness.set_limits(0, 100, incr_value=1)
        b.brightness.set_value(50)
        b.brightness.add_callback('value-changed', self.brightness_set_cb)
        b.brightness.set_tooltip("Set brightness/bias for the viewer")

        btn = Widgets.Button('Default Contrast')
        btn.set_tooltip("Reset contrast to default")
        btn.add_callback('activated', self.restore_contrast_cb)
        b._hbox_cb.add_widget(btn, stretch=0)
        btn = Widgets.Button('Default Brightness')
        btn.set_tooltip("Reset brightness to default")
        btn.add_callback('activated', self.restore_brightness_cb)
        b._hbox_cb.add_widget(btn, stretch=0)

        fr.set_widget(w)
        top.add_widget(fr, stretch=0)

        # add colorbar
        fr = Widgets.Frame("Color Map")
        height = 50
        settings = self.rgbmap.get_settings()
        settings.set(cbar_height=height, fontsize=10)
        cbar = ColorBar.ColorBar(self.logger, rgbmap=self.rgbmap,
                                 link=True,
                                 settings=settings)
        vmax = self.rgbmap.get_hash_size() - 1
        cbar.cbar_view.cut_levels(0, vmax)
        cbar_w = cbar.get_widget()
        cbar_w.resize(-1, height)

        self.colorbar = cbar
        #cbar.add_callback('motion', self.cbar_value_cb)

        top.add_widget(cbar_w, stretch=0)

        container.set_widget(top)

    @property
    def calg_name(self):
        return self._calg_name

    @calg_name.setter
    def calg_name(self, val):
        self._calg_name = val
        if self.gui_up:
            idx = self.calg_names.index(val)
            self.w.algorithm.set_index(idx)
            self.rgbmap.set_color_algorithm(val)

    @property
    def cmap_name(self):
        return self._cmap_name

    @cmap_name.setter
    def cmap_name(self, val):
        self._cmap_name = val
        if self.gui_up:
            idx = self.cmap_names.index(val)
            self.w.colormap.set_index(idx)
            self.rgbmap.set_color_map(val)

    @property
    def imap_name(self):
        return self._imap_name

    @imap_name.setter
    def imap_name(self, val):
        self._imap_name = val
        if self.gui_up:
            idx = self.imap_names.index(val)
            self.w.intensity.set_index(idx)
            self.rgbmap.set_intensity_map(val)

    @property
    def invert_cmap(self):
        return self._invert_cmap

    @invert_cmap.setter
    def invert_cmap(self, tf):
        self._invert_cmap = tf
        if self.gui_up:
            self.w.invert_cmap.set_state(tf)
            self.rgbmap.get_settings().set(color_map_invert=tf)

    @property
    def rotate_cmap(self):
        return self._rotate_cmap_pct

    @rotate_cmap.setter
    def rotate_cmap(self, pct):
        self._rotate_cmap_pct = pct
        if self.gui_up:
            self.w.rotate_cmap.set_value(int(pct * 100.0))
            self.rgbmap.get_settings().set(color_map_rot_pct=pct)

    @property
    def contrast(self):
        return self._contrast

    @contrast.setter
    def contrast(self, pct):
        self._contrast = pct
        if self.gui_up:
            self.w.contrast.set_value(int(pct * 100))
            self.rgbmap.get_settings().set(contrast=pct)

    @property
    def brightness(self):
        return self._brightness

    @brightness.setter
    def brightness(self, pct):
        self._brightness = pct
        if self.gui_up:
            self.w.brightness.set_value(int(pct * 100))
            self.rgbmap.get_settings().set(brightness=pct)

    def set_cmap_cb(self, w, index):
        """This callback is invoked when the user selects a new color
        map from the UI."""
        old_cmap_name = self._cmap_name
        name = mod_cmap.get_names()[index]
        self.cmap_name = name
        self.pipeline.push(StageAction(self,
                                       dict(cmap_name=old_cmap_name),
                                       dict(cmap_name=self._cmap_name),
                                       descr="rgbmap / change cmap"))

        self.pipeline.run_from(self)

    def set_imap_cb(self, w, index):
        """This callback is invoked when the user selects a new intensity
        map from the preferences pane."""
        old_imap_name = self._imap_name
        name = mod_imap.get_names()[index]
        self.imap_name = name
        self.pipeline.push(StageAction(self,
                                       dict(imap_name=old_imap_name),
                                       dict(imap_name=self._imap_name),
                                       descr="rgbmap / change imap"))

        self.pipeline.run_from(self)

    def set_calg_cb(self, w, index):
        """This callback is invoked when the user selects a new color
        hashing algorithm from the UI."""
        old_calg_name = self._calg_name
        name = self.calg_names[index]
        self.calg_name = name
        self.pipeline.push(StageAction(self,
                                       dict(calg_name=old_calg_name),
                                       dict(calg_name=self._calg_name),
                                       descr="rgbmap / change calg"))

        self.pipeline.run_from(self)

    def rotate_cmap_cb(self, w, val):
        old_val = self._rotate_cmap_pct
        pct = val / 100.0
        self.rotate_cmap = pct
        self.pipeline.push(StageAction(self,
                                       dict(rotate_cmap=old_val),
                                       dict(rotate_cmap=self._rotate_cmap_pct),
                                       descr=f"rgbmap / rotate cmap: {pct}"))

        self.pipeline.run_from(self)

    def invert_cmap_cb(self, w, tf):
        old_val = self._invert_cmap
        self.invert_cmap = tf
        self.pipeline.push(StageAction(self,
                                       dict(invert_cmap=old_val),
                                       dict(invert_cmap=self._invert_cmap),
                                       descr=f"rgbmap / invert cmap {tf}"))

        self.pipeline.run_from(self)

    def unrotate_cmap_cb(self, w):
        self.rotate_cmap_cb(w, 0)

    def set_default_cmaps(self):
        old = dict(cmap_name=self._cmap_name, imap_name=self._imap_name)
        cmap_name = "gray"
        imap_name = "ramp"
        new = dict(cmap_name=cmap_name, imap_name=imap_name)
        self.pipeline.push(StageAction(self, old, new,
                                       descr="rgbmap / change cmap,imap"))
        self.cmap_name = cmap_name
        self.imap_name = imap_name

        self.pipeline.run_from(self)

    def contrast_set_cb(self, w, val):
        old_val = self._contrast
        pct = val / 100.0
        self.contrast = pct
        self.pipeline.push(StageAction(self,
                                       dict(contrast=old_val),
                                       dict(contrast=self._contrast),
                                       descr=f"rgbmap / contrast: {pct}"))

        self.pipeline.run_from(self)

    def brightness_set_cb(self, w, val):
        old_val = self._brightness
        pct = val / 100.0
        self.brightness = pct
        self.pipeline.push(StageAction(self,
                                       dict(brightness=old_val),
                                       dict(brightness=self._brightness),
                                       descr=f"rgbmap / brightness: {pct}"))

        self.pipeline.run_from(self)

    def restore_contrast_cb(self, w):
        self.contrast_set_cb(w, 50)

    def restore_brightness_cb(self, w):
        self.brightness_set_cb(w, 50)

    def set_default_distmaps(self):
        old = dict(calg_name=self._calg_name)
        name = 'linear'
        new = dict(calg_name=name)
        self.pipeline.push(StageAction(self, old, new,
                                       descr="rgbmap / change calg"))
        self.calg_name = name

        self.pipeline.run_from(self)

    def copy_from_viewer_cb(self, w):
        rgbmap = self.viewer.get_rgbmap()
        rgbmap.copy_attributes(self.rgbmap, keylist=self.settings_keys)

        self.pipeline.run_from(self)

    def rgbmap_changed_cb(self, rgbmap):
        self.fv.gui_do_oneshot('pl-rgbmap', self.pipeline.run_from, self)

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        if self._bypass or data is None:
            self.pipeline.send(res_np=data)
            return

        if not np.issubdtype(data.dtype, np.uint):
            data = data.astype(np.uint)

        # get RGB mapped array
        res_np = self.rgbmap.get_rgb_array(data, order=self.order)

        self.pipeline.send(res_np=res_np)

    def __str__(self):
        return self._stagename

    def _get_state(self):
        return dict(calg_name=self._calg_name, cmap_name=self._cmap_name,
                    imap_name=self._imap_name)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.calg_name = d['calg_name']
        self.cmap_name = d['cmap_name']
        self.imap_name = d['imap_name']



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``PluginConfig`` plugin allows you to configure the plugins that
are visible in your menus.

**Plugin Type: Global**

``PluginConfig`` is a global plugin.  Only one instance can be opened.

**Usage**

PluginConfig is used to configure plugins to be used in Ginga.  The items
that can be configured for each plugin include:

* whether it is enabled (and therefore whether it shows up in the menus)
* the category of the plugin (used to construct the menu hierarchy)
* the workspace in which the plugin will open
* if a global plugin, whether it starts automatically when the reference
  viewer starts
* Whether the plugin name should be hidden (not show up in plugin
  activation menus)

When PluginConfig starts, it will show a table of plugins.  To edit the
above attributes for plugins, click "Edit", which will bring up a dialog
for editing the table.

For each plugin you want to configure, click on an entry in the main table
and then adjust the settings in the dialog, then click "Set" in the dialog
to reflect the changes back into the table.  If you don't click "Set",
nothing is changed in the table.  When you are done editing configurations,
click "Close" on the dialog to close the editing dialog.

.. note:: It is not recommended to change the workspace for a plugin
          unless you choose a compatibly-sized workspace to the original,
          as the plugin may not display correctly.  If in doubt, leave
          the workspace unchanged.  Also, disabling plugins in the
          "Systems" category may cause some expected features to stop
          working.


.. important:: To make the changes persist across Ginga restarts, click
               "Save" to save the settings (to `$HOME/.ginga/plugins.json`).
               Restart Ginga to see changes to the menus (via "category"
               changes).  **Remove this file manually if you want to reset
               the plugin configurations to the defaults**.


"""
import os.path

import yaml

from ginga import GingaPlugin
from ginga.util.paths import ginga_home
from ginga.gw import Widgets

__all__ = ['PluginConfig']


class PluginConfig(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        super().__init__(fv)

        self.plugin_dct = dict()

        self.columns = [("Name", 'name'),
                        ("Enabled", 'enabled'),
                        ("Type", 'ptype'),
                        ("Category", 'category'),
                        ("Workspace", 'workspace'),
                        ("Hidden", 'hidden'),
                        ("Auto Start", 'start')]

        self.gui_up = False

    def build_gui(self, container):
        vbox = Widgets.VBox()
        vbox.set_spacing(1)

        tv = Widgets.TreeView(sortable=True, use_alt_row_color=True,
                              selection='multiple')
        tv.add_callback('selected', self.select_cb)
        tv.setup_table(self.columns, 0, 'name')
        self.w.plugin_tbl = tv
        self.w.edit_dialog = None

        vbox.add_widget(tv, stretch=1)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(4)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Edit")
        btn.add_callback('activated', self.edit_plugin_selections_cb)
        btn.set_tooltip("Edit configuration of selected plugins")
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Save")
        btn.add_callback('activated', lambda w: self.save_plugins_cb())
        btn.set_tooltip("Save configuration of plugins\n"
                        "(restart Ginga to see changes to menus)")
        btns.add_widget(btn, stretch=0)

        btns.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(btns, stretch=0)
        container.add_widget(vbox, stretch=1)

        self.gui_up = True

    def start(self):
        if len(self.plugin_dct) == 0:
            # create plugin table if we haven't already done so
            dct = dict()
            for spec in self.fv.plugins:
                module = spec['module']
                klass = spec.get('klass', None)
                name = spec.get('name', module if klass is None else klass)
                enabled = spec.get('enabled', True)
                ptype = spec['ptype']
                start = 'False' if ptype == 'local' else str(spec.get('start', False))
                _dct = dict(name=name,
                            module=module,
                            enabled=str(enabled),
                            ptype=spec['ptype'],
                            category=spec.get('category', 'Custom'),
                            workspace=spec.get('workspace', 'in:dialog'),
                            hidden=str(spec.get('hidden', False)),
                            start=start)
                for key in ('menu', 'tab', 'klass', 'index', 'exclusive'):
                    if key in spec:
                        _dct[key] = spec[key]
                dct[name] = _dct
            self.plugin_dct = dct

        self.w.plugin_tbl.set_tree(self.plugin_dct)
        self.w.plugin_tbl.set_optimal_column_widths()

    def stop(self):
        self.gui_up = False

    def edit_plugin_selections_cb(self, w):
        # build dialog for editing
        captions = (("Enabled", 'checkbox'),
                    ("Category:", 'label', "category", 'textentry'),
                    ("Workspace:", 'label', "workspace", 'textentry'),
                    ("Auto start", 'checkbox', "Hidden", 'checkbox'),
                    )

        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)
        b.enabled.set_tooltip("Enable plugin(s)")
        b.enabled.set_enabled(False)
        b.category.set_tooltip("Edit menu category of plugin(s)")
        b.category.set_enabled(False)
        b.workspace.set_tooltip("Edit workspace of plugin(s)")
        b.workspace.set_enabled(False)
        b.auto_start.set_tooltip("Start plugin(s) at program startup\n"
                                 "(for global plugins only)")
        b.auto_start.set_enabled(False)
        b.hidden.set_tooltip("Hide plugin(s) names from program menus")
        b.hidden.set_enabled(False)

        dialog = Widgets.Dialog(title="Edit Plugin Configuration",
                                flags=0,
                                modal=False,
                                buttons=[['Set', 0], ['Close', 1]],
                                parent=self.w.plugin_tbl)
        dialog.add_callback('activated',
                            lambda w, rsp: self.edit_cb(w, rsp))
        self.w.edit_dialog = dialog

        box = dialog.get_content_area()
        box.set_border_width(4)
        box.add_widget(w, stretch=0)

        sel_dct = self.w.plugin_tbl.get_selected()
        self.select_cb(self.w.plugin_tbl, sel_dct)

        self.fv.ds.show_dialog(self.w.edit_dialog)

    def edit_cb(self, w, rsp):
        if rsp == 1:
            # close
            self.fv.ds.remove_dialog(w)
            self.w.edit_dialog = None
            return

        sel_dct = self.w.plugin_tbl.get_selected()
        if len(sel_dct) == 0:
            self.fv.show_error("No table entries selected",
                               raisetab=True)
            return

        enabled = self.w.enabled.get_state()
        category = self.w.category.get_text().strip()
        workspace = self.w.workspace.get_text().strip()
        hidden = self.w.hidden.get_state()
        start = self.w.auto_start.get_state()

        if len(category) == 0:
            self.fv.show_error("Category field should not be empty",
                               raisetab=True)
            return

        if len(workspace) == 0:
            self.fv.show_error("Workspace field should not be empty",
                               raisetab=True)
            return

        for name, pl_dct in sel_dct.items():
            self.plugin_dct[name]['enabled'] = str(enabled)
            self.plugin_dct[name]['category'] = category
            self.plugin_dct[name]['workspace'] = workspace
            if self.plugin_dct[name]['ptype'] == 'global':
                self.plugin_dct[name]['start'] = str(start)
            self.plugin_dct[name]['hidden'] = str(hidden)

        self.w.plugin_tbl.set_tree(self.plugin_dct)

    def save_plugins_cb(self):
        path = os.path.join(ginga_home, 'plugins.yml')
        _plugins = []
        for key, dct in self.plugin_dct.items():
            d = dct.copy()
            d['enabled'] = (d['enabled'] == 'True')
            if d['ptype'] == 'global':
                d['start'] = (d['start'] == 'True')
            else:
                d['start'] = False
            d['hidden'] = (d['hidden'] == 'True')
            _plugins.append(d)
        try:
            with open(path, 'w') as out_f:
                out_f.write(yaml.dump(_plugins))

        except Exception as e:
            self.logger.error(f"failed to save plugin file: {e}",
                              exc_info=True)
            self.fv.show_error(str(e))

    def select_cb(self, w, dct):
        if self.w.edit_dialog is None:
            return
        selected = len(dct) > 0
        self.w.enabled.set_enabled(selected)
        enabled = set([p_dct['enabled'] for p_dct in dct.values()])
        if len(enabled) == 1:
            self.w.enabled.set_state(enabled.pop() == 'True')
        else:
            self.w.enabled.set_state(True)

        self.w.category.set_enabled(selected)
        # only set category widget if all selected are the same
        # unique value
        categories = set([p_dct['category'] for p_dct in dct.values()])
        if len(categories) == 1:
            self.w.category.set_text(categories.pop())
        else:
            self.w.category.set_text('')

        self.w.workspace.set_enabled(selected)
        # only set workspace widget if all selected are the same
        # unique value
        workspaces = set([p_dct['workspace'] for p_dct in dct.values()])
        if len(workspaces) == 1:
            self.w.workspace.set_text(workspaces.pop())
        else:
            self.w.workspace.set_text('')

        # only enable auto start widget if all ptypes are 'global'
        ptypes = set([p_dct['ptype'] for p_dct in dct.values()])
        self.w.auto_start.set_enabled(len(ptypes) == 1 and 'global' in ptypes)

        starts = set([p_dct['start'] for p_dct in dct.values()])
        if len(starts) == 1:
            auto_start = (starts.pop() == 'True')
            self.w.auto_start.set_state(auto_start)
        else:
            self.w.auto_start.set_state(False)

        # only set hidden widget if all selected are hidden == 'True'
        self.w.hidden.set_enabled(selected)
        hiddens = set([p_dct['hidden'] for p_dct in dct.values()])
        if len(hiddens) == 1:
            hidden = (hiddens.pop() == 'True')
            self.w.hidden.set_state(hidden)
        else:
            self.w.hidden.set_state(False)

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'pluginconfig'



"""Module to handle image quality calculations using ``astropy``."""

import numpy as np
from astropy.modeling import models, fitting

from ginga.misc import Bunch

# Reuse shared items from the old module until we can merge old and new together.
from ginga.util.iqcalc import IQCalcError, get_median, have_scipy
from ginga.util.iqcalc import IQCalc as _IQCalc

# Import the rest into namespace so we can use this module like iqcalc.
from ginga.util.iqcalc import get_mean  # noqa

try:
    from photutils.centroids import centroid_com
    from photutils.detection import find_peaks
    have_photutils = True
except ImportError:
    have_photutils = False

__all__ = ['IQCalc']


# TODO: Use photutils for source finding.
class IQCalc(_IQCalc):
    """This is `ginga.util.iqcalc.IQCalc` that uses ``astropy``.

    This subclass has an extra ``self.fitter`` attribute for ``astropy``
    fitting.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fitter = fitting.LevMarLSQFitter()

    # FWHM CALCULATION

    def _prep_for_fitting(self, arr1d, medv):
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")

        N = len(arr1d)
        X = np.array(list(range(N)))
        Y = np.asarray(arr1d)

        # Fitting works more reliably if we do the following
        # a. subtract sky background
        if medv is None:
            medv = get_median(Y)
        Y = Y - medv
        maxv = Y.max()
        # b. clamp to 0..max (of the sky subtracted field)
        Y = Y.clip(0, maxv)

        return N, X, Y, maxv

    def gaussian(self, x, p):
        """Evaluate Gaussian function in 1D.

        Parameters
        ----------
        x : array-like
            X values.

        p : tuple of float
            Parameters for Gaussian, i.e., ``(mean, stddev, amplitude)``.

        Returns
        -------
        y : array-like
            Y values.

        """
        g = models.Gaussian1D(amplitude=p[2], mean=p[0], stddev=p[1])
        return g(x)

    def calc_fwhm_gaussian(self, arr1d, medv=None, **kwargs):
        """FWHM calculation on a 1D array by using least square fitting of
        a Gaussian function on the data.

        Parameters
        ----------
        arr1d : array-like
            1D array cut in either X or Y direction on the object.

        medv : float or `None`
            Median of the data. If not given, it is calculated from ``arr1d``.

        kwargs : dict
            Not used; for backward-compatible API call only.

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            Fitting results.

        Raises
        ------
        IQCalcError
            Fitting failed.

        """
        N, X, Y, maxv = self._prep_for_fitting(arr1d, medv)

        # Gaussian model with initial guess
        m_init = models.Gaussian1D(
            amplitude=maxv, mean=N * 0.5, stddev=(N * 0.25))

        # NOTE: without this mutex, optimize.leastsq causes a fatal error
        # sometimes--it appears not to be thread safe.
        # The error is:
        # "SystemError: null argument to internal routine"
        # "Fatal Python error: GC object already tracked"
        with self.lock:
            try:
                m = self.fitter(m_init, X, Y)
            except Exception:
                raise IQCalcError("FWHM Gaussian fitting failed")

        # Now that we have the sdev from fitting, we can calculate FWHM
        fwhm = m.fwhm
        # Some routines choke on numpy values and need "pure" Python floats
        # e.g. when marshalling through a remote procedure interface
        mu = m.mean.value
        sdev = m.stddev.value
        maxv = m.amplitude.value

        self.logger.debug('mu={} sdev={} maxv={}'.format(mu, sdev, maxv))

        res = Bunch.Bunch(fwhm=fwhm, mu=mu, sdev=sdev, maxv=maxv,
                          fit_fn=self.gaussian, fit_args=[mu, sdev, maxv])
        return res

    def moffat(self, x, p):
        """Evaluate Moffat function in 1D.

        Parameters
        ----------
        x : array-like
            X values.

        p : tuple of float
            Parameters for Moffat, i.e., ``(x_0, gamma, alpha, amplitude)``,
            where ``x_0`` a.k.a. mean and ``gamma`` core width.

        Returns
        -------
        y : array-like
            Y values.

        """
        m = models.Moffat1D(amplitude=p[3], x_0=p[0], gamma=p[1], alpha=p[2])
        return m(x)

    def calc_fwhm_moffat(self, arr1d, medv=None, **kwargs):
        """FWHM calculation on a 1D array by using least square fitting of
        a Moffat function on the data.

        Parameters
        ----------
        arr1d : array-like
            1D array cut in either X or Y direction on the object.

        medv : float or `None`
            Median of the data. If not given, it is calculated from ``arr1d``.

        kwargs : dict
            Not used; for backward-compatible API call only.

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            Fitting results.

        Raises
        ------
        IQCalcError
            Fitting failed.

        """
        N, X, Y, maxv = self._prep_for_fitting(arr1d, medv)

        # Moffat model with initial guess
        m_init = models.Moffat1D(
            amplitude=maxv, x_0=(N * 0.5), gamma=(N * 0.25), alpha=2)

        # NOTE: without this mutex, optimize.leastsq causes a fatal error
        # sometimes--it appears not to be thread safe.
        # The error is:
        # "SystemError: null argument to internal routine"
        # "Fatal Python error: GC object already tracked"
        with self.lock:
            try:
                m = self.fitter(m_init, X, Y)
            except Exception:
                raise IQCalcError("FWHM Moffat fitting failed")

        fwhm = m.fwhm

        # Some routines choke on numpy values and need "pure" Python floats
        # e.g. when marshalling through a remote procedure interface
        mu = m.x_0.value
        width = np.abs(m.gamma.value)
        power = m.alpha.value
        maxv = m.amplitude.value

        self.logger.debug('mu={} width={} power={} maxv={}'.format(
            mu, width, power, maxv))

        res = Bunch.Bunch(fwhm=fwhm, mu=mu, width=width, power=power,
                          maxv=maxv, fit_fn=self.moffat,
                          fit_args=[mu, width, power, maxv])
        return res

    def lorentz(self, x, p):
        """Evaluate Lorentz function in 1D.

        Parameters
        ----------
        x : array-like
            X values.

        p : tuple of float
            Parameters for Lorentz, i.e., ``(x_0, fwhm, amplitude)``.

        Returns
        -------
        y : array-like
            Y values.

        """
        m = models.Lorentz1D(amplitude=p[2], x_0=p[0], fwhm=p[1])
        return m(x)

    def calc_fwhm_lorentz(self, arr1d, medv=None, **kwargs):
        """FWHM calculation on a 1D array by using least square fitting of
        a Lorentz function on the data.

        Parameters
        ----------
        arr1d : array-like
            1D array cut in either X or Y direction on the object.

        medv : float or `None`
            Median of the data. If not given, it is calculated from ``arr1d``.

        kwargs : dict
            Not used; for backward-compatible API call only.

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            Fitting results.

        Raises
        ------
        IQCalcError
            Fitting failed.

        """
        N, X, Y, maxv = self._prep_for_fitting(arr1d, medv)

        # Lorentz model with initial guess
        m_init = models.Lorentz1D(
            amplitude=maxv, x_0=(N * 0.5), fwhm=(N * 0.25))

        # NOTE: without this mutex, optimize.leastsq causes a fatal error
        # sometimes--it appears not to be thread safe.
        # The error is:
        # "SystemError: null argument to internal routine"
        # "Fatal Python error: GC object already tracked"
        with self.lock:
            try:
                m = self.fitter(m_init, X, Y)
            except Exception:
                raise IQCalcError("FWHM Lorentz fitting failed")

        # Some routines choke on numpy values and need "pure" Python floats
        # e.g. when marshalling through a remote procedure interface
        fwhm = m.fwhm.value
        mu = m.x_0.value
        maxv = m.amplitude.value

        self.logger.debug('mu={} fwhm={} maxv={}'.format(mu, fwhm, maxv))

        res = Bunch.Bunch(fwhm=fwhm, mu=mu, maxv=maxv, fit_fn=self.lorentz,
                          fit_args=[mu, fwhm, maxv])
        return res

    def calc_fwhm(self, arr1d, medv=None, method_name='gaussian'):
        """Calculate FWHM for the given input array.

        Parameters
        ----------
        arr1d : array-like
            1D array cut in either X or Y direction on the object.

        medv : float or `None`
            Median of the data. If not given, it is calculated from ``arr1d``.

        method_name : {'gaussian', 'moffat', 'lorentz'}
            Function to use for fitting.

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            Fitting results.

        Raises
        ------
        NotImplementedError
            Given function is not supported.

        """
        if method_name == 'gaussian':
            fwhm_fn = self.calc_fwhm_gaussian
        elif method_name == 'moffat':
            fwhm_fn = self.calc_fwhm_moffat
        elif method_name == 'lorentz':
            fwhm_fn = self.calc_fwhm_lorentz
        else:
            raise NotImplementedError(
                'Fitting with {} is unsupported'.format(method_name))

        return fwhm_fn(arr1d, medv=medv)

    def centroid(self, data, xc, yc, radius):
        if not have_photutils:
            raise IQCalcError("Please install the 'photutils' package "
                              "to use this function")

        x0, y0, arr = self.cut_region(int(xc), int(yc), int(radius), data)
        cx, cy = centroid_com(np.asarray(arr))  # Return (X, Y), not (Y, X)
        return (x0 + cx, y0 + cy)

    def find_bright_peaks(self, data, threshold=None, sigma=5, radius=5):
        if not have_photutils:
            raise IQCalcError("Please install the 'photutils' package "
                              "to use this function")

        if threshold is None:
            # set threshold to default if none provided
            threshold = self.get_threshold(data, sigma=sigma)
            self.logger.debug(f"threshold defaults to {threshold} (sigma={sigma})")

        if np.ma.is_masked(data):
            mask = data.mask
            data = data.data
        else:
            mask = None

        out_tab = find_peaks(data, threshold, box_size=(radius * 2), mask=mask)
        peaks = list(zip(out_tab['x_peak'], out_tab['y_peak']))

        self.logger.debug(f"peaks={peaks}")
        return peaks

    # TODO: Perhaps evaluate_peaks or pick_field can also use photutils.
    # https://github.com/astropy/photutils/issues/1074



#
# image.py -- classes for images drawn on Ginga canvases.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import uuid
import weakref
import numpy as np

from ginga.canvas.CanvasObject import (CanvasObjectBase, _bool, _color,
                                       Point, MovePoint,
                                       register_canvas_types,
                                       colors_plus_none, coord_names)
from ginga.misc.ParamSet import Param
from ginga.misc import Bunch

from .mixins import OnePointMixin


class ImageP(OnePointMixin, CanvasObjectBase):
    """Draws an image on a Ginga canvas.
    Parameters are:
    x, y: 0-based coordinates of one corner in the data space
    image: the image, which must be an RGBImage object
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of corner of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of corner of object"),
            ## Param(name='image', type=?, argpos=2,
            ##       description="Image to be displayed on canvas"),
            Param(name='scale_x', type=float, default=1.0,
                  description="Scaling factor for X dimension of object"),
            Param(name='scale_y', type=float, default=1.0,
                  description="Scaling factor for Y dimension of object"),
            Param(name='interpolation', type=str, default=None,
                  description="Interpolation method for scaling pixels"),
            Param(name='linewidth', type=int, default=0,
                  min=0, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default: solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='lightgreen',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            Param(name='flipy', type=_bool,
                  default=False, valid=[False, True],
                  description="Flip image in Y direction"),
            Param(name='optimize', type=_bool,
                  default=True, valid=[False, True],
                  description="Optimize rendering for this object"),
        ]

    def __init__(self, pt, image, alpha=1.0, scale_x=1.0, scale_y=1.0,
                 interpolation=None,
                 linewidth=0, linestyle='solid', color='lightgreen',
                 showcap=False, flipy=False, optimize=True,
                 **kwdargs):
        self.kind = 'image'
        points = np.asarray([pt], dtype=float)
        CanvasObjectBase.__init__(self, points=points, image=image, alpha=alpha,
                                  scale_x=scale_x, scale_y=scale_y,
                                  interpolation=interpolation,
                                  linewidth=linewidth, linestyle=linestyle,
                                  color=color, showcap=showcap,
                                  flipy=flipy, optimize=optimize,
                                  **kwdargs)
        OnePointMixin.__init__(self)

        # The cache holds intermediate step results by viewer.
        # Depending on value of `whence` they may not need to be recomputed.
        self._cache = weakref.WeakKeyDictionary()
        self.image_id = str(uuid.uuid4())
        # images are not editable by default
        self.editable = False
        # is this image "data" or something else
        self.is_data = False

        self.enable_callback('image-set')

    def in_cache(self, viewer):
        return viewer in self._cache

    def get_cache(self, viewer):
        if viewer in self._cache:
            cache = self._cache[viewer]
        else:
            cache = self._reset_cache(Bunch.Bunch())
            self._cache[viewer] = cache
        return cache

    def invalidate_cache(self, viewer):
        cache = self.get_cache(viewer)
        self._reset_cache(cache)
        return cache

    def draw(self, viewer):
        """General draw method for RGB image types.
        Note that actual insertion of the image into the output is
        handled in `draw_image()`
        """
        if self.image is None:
            return
        whence = viewer._whence
        cache = self.get_cache(viewer)

        if not cache.drawn:
            # Normally, prepare_image() will be called earlier, and
            # cache.drawn will be true.  This catches the case for
            # some renderers where that might not happen
            viewer.prepare_image(self, cache, whence)

        cpoints = self.get_cpoints(viewer)
        cr = viewer.renderer.setup_cr(self)

        cr.draw_image(self, cpoints, cache, whence)

        # draw optional border
        if self.linewidth > 0:
            cr.draw_polygon(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)

    def prepare_image(self, viewer, whence):
        cache = self.get_cache(viewer)
        viewer.prepare_image(self, cache, whence)

    def _reset_cache(self, cache):
        cache.setvals(cutout=None, rgbarr=None, drawn=False, visible=False,
                      cvs_pos=(0, 0))
        return cache

    def reset_optimize(self):
        for cache in self._cache.values():
            self._reset_cache(cache)

    def get_image(self):
        return self.image

    def set_image(self, image):
        self.image = image
        self.reset_optimize()

        self.make_callback('image-set', image)

    def get_scaled_wdht(self):
        if self.image is None:
            return (0, 0)
        width = self.image.width * self.scale_x
        height = self.image.height * self.scale_y
        return (width, height)

    def get_coords(self):
        x1, y1 = self.crdmap.to_data((self.x, self.y))
        # TODO: this should be viewer.data_off instead of hard-coded,
        # but we don't have a handle to the viewer here.
        x1, y1 = x1 - 0.5, y1 - 0.5
        wd, ht = self.get_scaled_wdht()
        x2, y2 = x1 + wd, y1 + ht
        return (x1, y1, x2, y2)

    def get_llur(self):
        return self.get_coords()

    def get_center_pt(self):
        x1, y1, x2, y2 = self.get_coords()
        return ((x1 + x2) * 0.5, (y1 + y2) * 0.5)

    def get_points(self):
        x1, y1, x2, y2 = self.get_coords()
        return [(x1, y1), (x2, y1), (x2, y2), (x1, y2)]

    def contains_pts(self, pts):
        if len(pts) == 0:
            return np.array([], dtype=bool)
        x_arr, y_arr = np.asarray(pts).T
        x1, y1, x2, y2 = self.get_coords()

        contains = np.logical_and(
            np.logical_and(x1 <= x_arr, x_arr < x2),
            np.logical_and(y1 <= y_arr, y_arr < y2))
        return contains

    def rotate(self, theta, xoff=0, yoff=0):
        raise ValueError("Images cannot be rotated")

    def setup_edit(self, detail):
        detail.center_pos = self.get_center_pt()
        detail.scale_x = self.scale_x
        detail.scale_y = self.scale_y

    def set_edit_point(self, i, pt, detail):
        if i == 0:
            self.move_to_pt(pt)
        elif i == 1:
            scale_x = self.calc_scale_from_pt(pt, detail)
            self.scale_x = detail.scale_x * scale_x
        elif i == 2:
            scale_y = self.calc_scale_from_pt(pt, detail)
            self.scale_y = detail.scale_y * scale_y
        elif i == 3:
            scale_x, scale_y = self.calc_dual_scale_from_pt(pt, detail)
            self.scale_x = detail.scale_x * scale_x
            self.scale_y = detail.scale_y * scale_y
        else:
            raise ValueError("No point corresponding to index %d" % (i))

        self.reset_optimize()
        detail.viewer.redraw(whence=0)

    def get_edit_points(self, viewer):
        x1, y1, x2, y2 = self.get_coords()
        return [MovePoint(*self.get_center_pt()),    # location
                Point(x2, (y1 + y2) * 0.5),          # width scale
                Point((x1 + x2) * 0.5, y2),          # height scale
                Point(x2, y2),                       # both scale
                ]

    def scale_by_factors(self, factors):
        scale_x, scale_y = factors[:2]
        self.scale_x *= scale_x
        self.scale_y *= scale_y
        self.reset_optimize()

    def scale_by(self, scale_x, scale_y):
        # TO BE DEPRECATED--use scale_by_factors instead
        self.scale_by_factors((scale_x, scale_y))

    def set_scale(self, scale_x, scale_y):
        self.scale_x = scale_x
        self.scale_y = scale_y
        self.reset_optimize()

    def set_origin(self, x, y):
        self.x, self.y = x, y
        self.reset_optimize()


class Image(ImageP):

    def __init__(self, x, y, image, alpha=1.0, scale_x=1.0, scale_y=1.0,
                 interpolation=None,
                 linewidth=0, linestyle='solid', color='lightgreen',
                 showcap=False, flipy=False, optimize=True,
                 **kwdargs):
        ImageP.__init__(self, (x, y), image, alpha=alpha,
                        scale_x=scale_x, scale_y=scale_y,
                        interpolation=interpolation,
                        linewidth=linewidth, linestyle=linestyle,
                        color=color, showcap=showcap,
                        flipy=flipy, optimize=optimize,
                        **kwdargs)


class NormImageP(ImageP):
    """Draws an image on a Ginga canvas.

    Parameters are:
    x, y: 0-based coordinates of one corner in the data space
    image: the image, which must be an RGBImage object
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of corner of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of corner of object"),
            ## Param(name='image', type=?, argpos=2,
            ##       description="Image to be displayed on canvas"),
            Param(name='scale_x', type=float, default=1.0,
                  description="Scaling factor for X dimension of object"),
            Param(name='scale_y', type=float, default=1.0,
                  description="Scaling factor for Y dimension of object"),
            Param(name='interpolation', type=str, default=None,
                  description="Interpolation method for scaling pixels"),
            Param(name='linewidth', type=int, default=0,
                  min=0, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default: solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='lightgreen',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            ## Param(name='flipy', type=_bool,
            ##       default=False, valid=[False, True],
            ##       description="Flip image in Y direction"),
            Param(name='optimize', type=_bool,
                  default=True, valid=[False, True],
                  description="Optimize rendering for this object"),
            ## Param(name='cuts', type=tuple, default=None,
            ##       description="Tuple of (lo, hi) cut levels for image"),
            ## Param(name='rgbmap', type=?,
            ##       description="RGB mapper for the image"),
            ## Param(name='autocuts', type=?,
            ##       description="Cuts manager for the image"),
        ]

    def __init__(self, pt, image, alpha=1.0, scale_x=1.0, scale_y=1.0,
                 interpolation=None, cuts=None, linewidth=0, linestyle='solid',
                 color='lightgreen', showcap=False,
                 optimize=True, rgbmap=None, autocuts=None, **kwdargs):
        super(NormImageP, self).__init__(pt, image, alpha=alpha,
                                         scale_x=scale_x, scale_y=scale_y,
                                         interpolation=interpolation,
                                         linewidth=linewidth, linestyle=linestyle,
                                         color=color,
                                         showcap=showcap, optimize=optimize,
                                         **kwdargs)
        self.kind = 'normimage'
        self.rgbmap = rgbmap
        self.cuts = cuts
        self.autocuts = autocuts

    def _reset_cache(self, cache):
        cache.setvals(cutout=None, alpha=None, prergb=None, rgbarr=None,
                      drawn=False, visible=False, cvs_pos=(0, 0))
        return cache


class NormImage(NormImageP):

    def __init__(self, x, y, image, alpha=1.0, scale_x=1.0, scale_y=1.0,
                 interpolation=None, cuts=None, linewidth=0, linestyle='solid',
                 color='lightgreen', showcap=False,
                 optimize=True, rgbmap=None, autocuts=None, **kwdargs):
        NormImageP.__init__(self, (x, y), image, alpha=alpha,
                            scale_x=scale_x, scale_y=scale_y,
                            interpolation=interpolation,
                            linewidth=linewidth, linestyle=linestyle,
                            color=color, showcap=showcap, optimize=optimize,
                            **kwdargs)


# register our types
register_canvas_types(dict(image=Image, normimage=NormImage))

#END



#
# CanvasRenderVec.py -- for rendering into a vector of drawing operations
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

from .VecHelp import (IMAGE, LINE, CIRCLE, BEZIER, ELLIPSE_BEZIER, POLYGON,
                      PATH, TEXT)

from ginga.canvas import render


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)

    def text_extents(self, text, font=None):
        if font is None:
            font = self.font
        return self.renderer.text_extents(text, font)

    ##### DRAWING OPERATIONS #####

    def draw_image(self, image_id, cpoints, rgb_arr, whence, order='RGB'):
        self.renderer.rl.append((IMAGE, (image_id, cpoints, rgb_arr, whence,
                                         order)))

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, fill=None,
                  line=None):
        self.renderer.rl.append((TEXT, (cx, cy, text, rot_deg),
                                 line, fill, font))

    def draw_polygon(self, cpoints, line=None, fill=None):
        self.renderer.rl.append((POLYGON, cpoints, line, fill))

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        self.renderer.rl.append((CIRCLE, (cx, cy, cradius),
                                 line, fill))

    ## def draw_bezier_curve(self, cpoints, line=None):
    ##     self.renderer.rl.append((BEZIER, cpoints, line))

    ## def draw_ellipse_bezier(self, cpoints, line=None, fill=None):
    ##     # draw 4 bezier curves to make the ellipse
    ##     self.renderer.rl.append((ELLIPSE_BEZIER, cpoints, line, fill))

    ## def draw_ellipse(self, cx, cy, cxradius, cyradius, rot_deg,
    ##                  line=None, fill=None):
    ##     self.renderer.rl.append((ELLIPSE,
    ##                              (cx, cy, cxradius, cyradius, rot_deg),
    ##                              line, fill))

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        self.renderer.rl.append((LINE, (cx1, cy1, cx2, cy2), line))

    def draw_path(self, cpoints, line=None):
        self.renderer.rl.append((PATH, cpoints, line))


class VectorRenderMixin:

    def __init__(self):
        # the render list
        self.rl = []

    def initialize(self):
        wd, ht = self.dims
        cpoints = ((0, 0), (wd, 0), (wd, ht), (ht, 0))
        bg = self.viewer.get_bg()
        line = self.get_line(color=bg)
        fill = self.get_fill(color=bg, alpha=1.0)
        self.rl = [(POLYGON, cpoints, line, fill)]

    def draw_vector(self, cr):
        for tup in self.rl:
            dtyp, line, fill, font = None, None, None, None
            try:
                dtyp = tup[0]
                if dtyp == IMAGE:
                    (image_id, cpoints, rgb_arr, whence, order) = tup[1]
                    cr.draw_image(image_id, cpoints, rgb_arr, whence,
                                  order=self.rgb_order)

                elif dtyp == LINE:
                    (cx1, cy1, cx2, cy2) = tup[1]
                    line = tup[2]
                    cr.draw_line(cx1, cy1, cx2, cy2, line=line)

                elif dtyp == CIRCLE:
                    cx, cy, cradius = tup[1]
                    line, fill = tup[2:4]
                    cr.draw_circle(cx, cy, cradius, line=line, fill=fill)

                elif dtyp == BEZIER:
                    cpoints = tup[1]
                    line = tup[2]
                    cr.draw_bezier_curve(cpoints, line=line)

                elif dtyp == ELLIPSE_BEZIER:
                    cpoints = tup[1]
                    line, fill = tup[2:4]
                    cr.draw_ellipse_bezier(cpoints, line=line, fill=fill)

                elif dtyp == POLYGON:
                    cpoints = tup[1]
                    line, fill = tup[2:4]
                    cr.draw_polygon(cpoints, line=line, fill=fill)

                elif dtyp == PATH:
                    cpoints = tup[1]
                    line = tup[2]
                    cr.draw_path(cpoints, line=line)

                elif dtyp == TEXT:
                    (cx, cy, text, rot_deg) = tup[1]
                    line, fill, font = tup[2:5]
                    cr.draw_text(cx, cy, text, rot_deg=rot_deg,
                                 font=font, fill=fill, line=line)

            except Exception as e:
                self.logger.error("Error drawing '{}': {}".format(dtyp, e),
                                  exc_info=True)

#END



import logging

import numpy as np

from ginga import AstroImage
from ginga.pilw.ImageViewPil import CanvasView


class TestImageView(object):

    def setup_class(self):
        self.logger = logging.getLogger("TestImageView")
        self.viewer = CanvasView(logger=self.logger)
        self.data = np.identity(2000)
        self.image = AstroImage.AstroImage(logger=self.logger)
        self.image.set_data(self.data)

    def test_scale(self):
        viewer = self.viewer
        viewer.set_window_size(900, 1100)
        viewer.set_image(self.image)
        zoom = 0.0
        scale_x = scale_y = 1.0
        viewer.scale_to(scale_x, scale_y)
        zoomlevel = viewer.get_zoom()
        assert zoomlevel == zoom

    def test_centering(self):
        viewer = self.viewer
        viewer.set_window_size(900, 1100)
        viewer.set_image(self.image)
        viewer.center_image()
        ht, wd = self.data.shape[:2]
        ctr_x, ctr_y = wd / 2. - viewer.data_off, ht / 2. - viewer.data_off
        pan_x, pan_y = viewer.get_pan()
        assert np.isclose(pan_x, ctr_x) and np.isclose(pan_y, ctr_y)

    def test_pan(self):
        viewer = self.viewer
        viewer.set_window_size(900, 1100)
        viewer.set_image(self.image)
        viewer.set_pan(401.0, 501.0)
        pan_x, pan_y = viewer.get_pan()
        assert np.isclose(pan_x, 401.0) and np.isclose(pan_y, 501.0)

    def test_pan2(self):
        viewer = self.viewer
        viewer.set_window_size(400, 300)
        viewer.set_image(self.image)
        viewer.set_pan(401.0, 501.0)
        viewer.scale_to(8.0, 8.0)
        x1, y1, x2, y2 = viewer.get_data_rect()
        result = np.array([(x1, y1), (x2, y2)])
        expected = np.array([[376., 482.25], [426., 519.75]])
        assert np.all(np.isclose(expected, result))
